﻿/**
モニタ。

History:
	$(HIST_VER 1.120,
		$(HIST_CHG module, monitor, display)
		$(HIST_MSG ↑関連部分の変更。)
	)
	$(HIST_VER 1.061,
		$(HIST_S package変更。)
	)
*/
module etc.negui.environ.display;

debug import std.stdio: wl = writefln;

debug(display) void main() {}

import std.exception;

import win32.windows;

import etc.negui.system.base;
import etc.negui.window.window;
import etc.negui.draw.canvas;

///
class DisplayException: NeGuiException {
	mixin MixInNeGuiException;
}

/**
History:
	$(HIST_VER 1.120,
		$(HIST_CHG 名前, Monitor, Display)
		$(HIST_B Object.Monitorと衝突。)
	)
*/
class Display: IHandle {
	private HMONITOR hMonitor;
	override const HANDLE opCall() {
		return hMonitor;
	}
	mixin MixInHandleRaiiNotDuplicate;
	mixin MixInHandleRaiiNotDup;

	/**
	*/
	enum FAILURE {
		NEAR      = MONITOR_DEFAULTTONEAREST, /// 指定したウィンドウに最も近い位置にあるディスプレイモニタのハンドルが返ります。
		PRIMARY   = MONITOR_DEFAULTTOPRIMARY, /// プライマリディスプレイモニタのハンドルが返ります。
		EXCEPTION = MONITOR_DEFAULTTONULL,    /// NULL が返ります。
	}
	/**
	*/
	this(HMONITOR hMonitor) {
		this.hMonitor = hMonitor;
	}
	/**
	*/
	this(in Window FromWindow, FAILURE Failure) {
		enforce(FromWindow, new DisplayException(Text("FromWindow is null")));
		
		auto hMonitor=MonitorFromWindow(FromWindow(), Failure);

		if(Failure == FAILURE.EXCEPTION) {
			enforce(hMonitor, new DisplayException(Text("null")));
		}

		this(hMonitor);
	}
	/**
	*/
	this(const ref POINT FromPoint, FAILURE Failure) {
		auto hMonitor=MonitorFromPoint(cast(POINT)FromPoint, Failure);

		if(Failure == FAILURE.EXCEPTION) {
			enforce(hMonitor, new DisplayException(Text("null")));
		}

		this(hMonitor);
	}
	/**
	*/
	this(const ref RECT FromRect, FAILURE Failure) {
		auto hMonitor=MonitorFromRect(cast(RECT*)&FromRect, Failure);

		if(Failure == FAILURE.EXCEPTION) {
			enforce(hMonitor, new DisplayException(Text("null")));
		}

		this(hMonitor);
	}

	/**
	History:
		$(HIST_VER 1.061,
			$(HIST_P 失敗時に例外。)
		)
	*/
	const ref DISPLAY information() {
		auto display=new DISPLAY;
		display.initialize;
		enforce(GetMonitorInfo(hMonitor, display.ptr), new DisplayException(ERR.toText));
		assert(display);
		return *display;
	}

	alias bool delegate(Display display, Canvas canvas, ref RECT Rect) EnumDisplayDelegate;
	private {
		struct ENUMDISPLAYCLOSURE {
			EnumDisplayDelegate dg;
			size_t callIndex;
		}
		static extern(Windows) BOOL MonitorEnumProc(HMONITOR hMonitor, HDC hdcMonitor, RECT *lprcMonitor, LPARAM dwData) {
			auto EnumDisplayClosure=cast(ENUMDISPLAYCLOSURE*)dwData;
			assert(EnumDisplayClosure);
			EnumDisplayClosure.callIndex++;
			auto display=new Display(hMonitor);
			auto canvas =new Canvas(hdcMonitor, false);
			return EnumDisplayClosure.dg(display, canvas, cast(RECT)*lprcMonitor);
		}
	}
	static bool enumDisplay(EnumDisplayDelegate dg) {
		ENUMDISPLAYCLOSURE EnumDisplayClosure;
		EnumDisplayClosure.dg = dg;
		if(EnumDisplayMonitors(null, null, &MonitorEnumProc, cast(LPARAM)&EnumDisplayClosure)) {
			return true;
		}
		return false;
	}
	static Display[] displays() {
		Display[] list;

		auto dg = cast(EnumDisplayDelegate)(Display display, Canvas canvas, ref RECT Rect) {
			list ~= display;
			return true;
		};
		enumDisplay(dg);
		
		return list;
	}
	static Display[] noPrimaryDisplays() {
		Display[] list;

		auto dg = cast(EnumDisplayDelegate)(Display display, Canvas canvas, ref RECT Rect) {
			if(display.information.flag != DISPLAY.FLAG.PRIMARY) {
				list ~= display;
			}
			return true;
		};
		enumDisplay(dg);
		
		return list;
	}

	override bool opEquals(Object o) {
		if(auto display=cast(Display)o) {
			return hMonitor == display.hMonitor;
		}
		return false;
	}
}


/***/
struct DISPLAY {
	MONITORINFOEX MonitorInfo;
	mixin(SMixInStructHiddenOriginal!(MONITORINFOEX)(q{MonitorInfo}));

	/**
	*/
	void initialize() {
		MonitorInfo.cbSize = MonitorInfo.sizeof;
	}
	/**
	*/
	ref RECT desktop() {
		return MonitorInfo.rcMonitor;
	}
	/**
	*/
	ref RECT work() {
		return MonitorInfo.rcWork;
	}
	/**
	*/
	enum FLAG {
		NONE = 0,
		PRIMARY = MONITORINFOF_PRIMARY,
	}
	mixin(SMixInStructGetSet!(FLAG)("flag", q{MonitorInfo.dwFlags}));

	/**
	*/
	const Text device() {
		return Text(cast(wchar*)MonitorInfo.szDevice.ptr);
	}
}





