﻿/**
ファイル・パス拡張操作。


History:
	$(HIST_VER 1.130,
		$(HIST_NEW)
	)
*/
module etc.negui.extra.file;

debug import etc.negui.nestdio: wl = writefln;
//debug(file) void main() {}

import std.exception;

import win32.shellapi;

import etc.negui.system.base;
import etc.negui.negui;
import etc.negui.file.file;
import etc.negui.file.exever;
import etc.negui.file.com.shortcut;
import etc.negui.system.type.enumerated;
import etc.negui.environ.variable;

/**
ファイル名の取得。

$(LIST_UNORDERED
	$(LIST_ITEM ファイルが実行形式でバージョン情報を持っているならその製品名。)
	$(LIST_ITEM 通常ファイルならその拡張子を取り除いた名前。)
	$(LIST_ITEM フォルダであれば最後の$(STRONG \)を取り除いた名前。)
	$(LIST_ITEM ドライブであればそのボリューム名。)
)

Params:
	Path = ファイル名を取得するアドレス。

Returns:
	それっぽい名前。

ToDo:
	ボリューム取得までのキチンとしたドライブ判定。

*/
ref Text GetName(in Text Path) {
	if(!Path.length) {
		return Text.init;
	}
	
	if(FILE.isExistence(Path)) {
		auto Name=new Text;
		if(FILE.isFile(Path)) {
			try {
				scope ver=new Version(Path);
				*Name = ver.fileInfo(Version.TYPE.PRODUCTNAME, ver.langAndCode[0]);
			} catch(VersionException) {
				assert(true, "処理継続");
			}
			if(Name.length) {
				return *Name;
			}
		} else {
			immutable IsDriveName=Path.data.length == `C:\`.length && Path.data[1] == ':';
			if(IsDriveName && IsIn!(PATH.cmp, const(Text))(Path, DRIVE.names())) {
				try {
					*Name=DRIVE.volumeInformation(Path).name;
					if(Name.length) {
						return *Name;
					}
				} catch(DriveException e) {
					assert(true, "処理継続");
				}
			}
			if(IsDriveName) {
				return PATH.getFileName(Path);
			}
			return PATH.delFolderSep(PATH.getFileName(Path));
		}
	} else if(PATH.isFolder(Path)) {
		return PATH.delFolderSep(PATH.getFileName(Path));
	}
	
	return PATH.getName(Path);
}

/**
ショートカット作成。

etc.negui.file.com.shortcut.Shortcutを明示的に使用せず手軽にショートカットを作成。

Params:
	SourcePath = ショートカット先アドレス。
	             完全なアドレスであること。

	ParentFolder = ショートカット生成先フォルダ。
	               このフォルダにGetNameで取得した名前のショートカットが置かれる。

	WorkFolder = ショートカットに設定される作業フォルダ。
	             空の場合SourcePathがファイルの場合はSourcePathの親フォルダが使用される。
	             SourcePathがフォルダの場合は無視される。

	Option = ショートカットに設定するコマンドライン引数。
	         SourcePathがフォルダの場合は無視される。

Returns:
	成功の真偽値。
	アドレスが不正、ショートカット生成先にすでにファイルが存在する場合等にこの関数は失敗する。

Note:
	COMの初期化が必要なため
	$(LIST_UNORDERED
		$(LIST_ITEM etc.negui.initializer.NeGuiInitialize)
		$(LIST_ITEM etc.negui.initializer.main)
		$(LIST_ITEM etc.negui.initializer.WinMain)
	)
	上記を使用していない場合はetc.negui.com.com.ComInitializeを呼び出す必要あり。
*/
bool MakeShortcut(in Text SourcePath, in Text ParentFolder, in Text WorkFolder=Text.init, in Text Option=Text.init) {
	if(!SourcePath.length) {
		// 文字列がなければそもそも作れない
		return false;
	}
	
	scope sc=new Shortcut();
	
	try {
		// ファイルでもフォルダでもなければ処理終了
		// エクスプローラの新規作成と挙動が違うけど問題なし。
		if(FILE.isFile(SourcePath)) {
			// ファイルであれば作業フォルダ、引数の設定。
			
			// 作業フォルダの設定、有効無効は問わない
			if(!WorkFolder.length) {
				sc.workFolder = PATH.ownerFolder(SourcePath);
			} else {
				sc.workFolder = WorkFolder;
			}
			
			// 引数
			sc.option = Option;
		}
		sc.address = SourcePath;
		
		if(!FILE.isFolder(ParentFolder)) {
			// フォルダじゃなければさようなら
			return false;
		}
	} catch(Exception e) {
		// 作れない
		return false;
	}

	// アイコンはアドレスと同じ。
	sc.setIcon(SourcePath, 0);
	// 通常表示
	sc.show = SHOW.SHOWNORMAL;

	// ファイル名生成
	auto FileName=GetName(SourcePath);
	if(!PATH.isName(FileName)) {
		FileName = PATH.delFolderSep(PATH.getName(SourcePath));
	}
	assert(PATH.isName(FileName));

	// アドレス生成
	auto ShortcutPath=PATH.addExtension(PATH.join(ParentFolder, FileName), Text("lnk"));

	if(FILE.isExistence(ShortcutPath)) {
		// 存在していれば保存しない
		return false;
	}
	
	return sc.fileSave(ShortcutPath, false);
}

version(none)
debug(file) {
	import etc.negui.com.com;
	unittest {
		ComInitialize();
	
		auto a=MakeShortcut(
			Text(`C:\Program Files\Internet Explorer\iexplore.exe`),
			Text(`Z:\temp\`),
			Text(``)
		);
		wl("%s", a);
	}
}

