﻿/**
文字列拡張操作。


History:
	$(HIST_VER 1.130,
		$(HIST_NEW)
	)
*/
module etc.negui.extra.text;

debug import etc.negui.nestdio: wl = writefln;
//debug(text) void main() {}

import std.exception;

import etc.negui.system.base;


/**
文字列置き換え。

文字列から連想配列のキーに一致する部分をそのキーに対応する値に置き換える。

Params:
	Source = 置き換え元の文字列。
	置き換え対象は$(STRONG ${.+})でありエスケープシーケンスは$(EMPHASIS $)。

	Map = 置き換えする文字列とその対象の文字列のペア。

Returns:
	置き換え後の文字列。

	Examples:
	* ----------------------------------------
	* Text[ConsText] map;
	* map[Text("a")] = Text("A");
	* map[Text("d")] = Text("D");
	* 
	* assert(ReplaceTextMap(Text("$${a}"), map) == Text("${a}"));
	* assert(ReplaceTextMap(Text("${a}"), map) == Text("A"));
	* assert(ReplaceTextMap(Text("$d}"), map) == Text("$d}"));
	* assert(ReplaceTextMap(Text("${}"), map) == Text("${}"));
	* assert(ReplaceTextMap(Text("${A}"), map) == Text("${A}"));
	* ----------------------------------------

*/
ConsText ReplaceTextMap(in Text Source, in Text[ConsText] Map) {
	if(!Map.length) {
		return Source;
	}
	if(!Source.length) {
		return Text.init;
	}

	const(Text) Replace(const ref Text Word, out bool IsChanged) {
		if(auto pKey=Word in Map) {
			IsChanged = true;
			return *pKey;
		} else {
			IsChanged = false;
			return Text.init;
		}
	}


	return Source.replaceToWord(Text("${"), '}', '$', &Replace);
}
debug(text) unittest {
	Text[ConsText] map;
	map[Text("a")] = Text("A");
	map[Text("b")] = Text("B");
	map[Text("c")] = Text("C");
	map[Text("d")] = Text("D");
	
	assert(ReplaceTextMap(Text("$${a}"), map) == Text("${a}"));
	assert(ReplaceTextMap(Text("${a}"), map) == Text("A"));
	assert(ReplaceTextMap(Text("${b}"), map) == Text("B"));
	assert(ReplaceTextMap(Text("${c}"), map) == Text("C"));
	assert(ReplaceTextMap(Text("${d}"), map) == Text("D"));
	assert(ReplaceTextMap(Text("$d}"), map) == Text("$d}"));
	assert(ReplaceTextMap(Text("${}"), map) == Text("${}"));
	assert(ReplaceTextMap(Text("${A}"), map) == Text("${A}"));
	assert(ReplaceTextMap(Text("${NONE}"), map) == Text("${NONE}"));
	assert(ReplaceTextMap(Text("${a}${b}${c}${d}"), map) == Text("ABCD"));
}




