﻿/**
バージョン情報用。
*/
module etc.negui.file.exever;
debug import std.stdio: wl = writefln;
debug(exever) void main() {}

import std.exception;

import win32.windows;
import win32.winver;

import etc.negui.system.base;
import etc.negui.file.file;

pragma(lib, "Version.Lib");

///
class VersionException: NeGuiException {
	mixin MixInNeGuiException;
}

/**
History:
	$(HIST_VER 1.061,
		$(HIST_S IText継承。)
	)
*/
class Version: IText {
	/**
	*/
	enum TYPE {
		COMMENTS,         /// 
		INTERNALNAME,     /// 
		PRODUCTNAME,      /// 
		COMPANYNAME,      /// 
		LEGALCOPYRIGHT,   /// 
		PRODUCTVERSION,   /// 
		FILEDESCRIPTION,  /// 
		LEGALTRADEMARKS,  /// 
		PRIVATEBUILD,     /// 
		FILEVERSION,      /// 
		ORIGINALFILENAME, /// 
		SPECIALBUILD,     /// 
	}
	//きもい
	private static struct DEF {
		static immutable wstring COMMENTS          = "Comments";
		static immutable wstring INTERNALNAME      = "InternalName";
		static immutable wstring PRODUCTNAME       = "ProductName";
		static immutable wstring COMPANYNAME       = "CompanyName";
		static immutable wstring LEGALCOPYRIGHT    = "LegalCopyright";
		static immutable wstring PRODUCTVERSION    = "ProductVersion";
		static immutable wstring FILEDESCRIPTION   = "FileDescription";
		static immutable wstring LEGALTRADEMARKS   = "LegalTrademarks";
		static immutable wstring PRIVATEBUILD      = "PrivateBuild";
		static immutable wstring FILEVERSION       = "FileVersion";
		static immutable wstring ORIGINALFILENAME  = "OriginalFilename";
		static immutable wstring SPECIALBUILD      = "SpecialBuild";

		pure nothrow static wstring typeToDef(TYPE Type) {
			final switch(Type) {
				case TYPE.COMMENTS:         return DEF.COMMENTS;
				case TYPE.INTERNALNAME:     return DEF.INTERNALNAME;
				case TYPE.PRODUCTNAME:      return DEF.PRODUCTNAME;
				case TYPE.COMPANYNAME:      return DEF.COMPANYNAME;
				case TYPE.LEGALCOPYRIGHT:   return DEF.LEGALCOPYRIGHT;
				case TYPE.PRODUCTVERSION:   return DEF.PRODUCTVERSION;
				case TYPE.FILEDESCRIPTION:  return DEF.FILEDESCRIPTION;
				case TYPE.LEGALTRADEMARKS:  return DEF.LEGALTRADEMARKS;
				case TYPE.PRIVATEBUILD:     return DEF.PRIVATEBUILD;
				case TYPE.FILEVERSION:      return DEF.FILEVERSION;
				case TYPE.ORIGINALFILENAME: return DEF.ORIGINALFILENAME;
				case TYPE.SPECIALBUILD:     return DEF.SPECIALBUILD;
			}
		}
	}
	private {
		immutable size_t DataSize;
		immutable void[] VersionData;
	}
	/**
	*/
	this(in Text FilePath) {
		try {
			enforce(FILE.isFile(FilePath), new VersionException(Text("%s is Folder", FilePath)));

			DWORD Handle;
			DataSize = GetFileVersionInfoSize(FilePath.ptr, &Handle);
			enforce(DataSize, new VersionException(ERR.toText));
			
			scope TempData = new void[DataSize];
			enforce(GetFileVersionInfo(FilePath.ptr, Handle, DataSize, TempData.ptr), new VersionException(ERR.toText));
			VersionData = TempData.idup;
		} catch(Exception e) {
			throw new VersionException(Text(this.classinfo.name ~ ".constructor"), e);
		}
	}
	/**
	*/
	struct LANGANDCODEPAGE {
		WORD Language;
		WORD CodePage;
		const ConsText toText() {
			return Text("%04x%04x", Language, CodePage);
		}
	}
	private const size_t Query(in Text SubBlock, void* Data) {
		size_t DataLength;
		
		if(VerQueryValue(cast(void*)VersionData.ptr, SubBlock.ptr, cast(void**)Data, &DataLength)) {
			return DataLength;
		}

		return 0;
	}
	/**
	*/
	const immutable(LANGANDCODEPAGE[]) langAndCode() {
		LANGANDCODEPAGE* LangCode;
		auto Size=Query(Text(`\VarFileInfo\Translation`), &LangCode);
		return cast(immutable)LangCode[0..Size];
	}
	
	private const Text FileInfo(in wstring Define, in LANGANDCODEPAGE LangAndCode) {
		const SubBlock=Text(`\StringFileInfo\%s\%s`, LangAndCode.toText, Define);
		
		wchar* Data;
		auto DataLength=Query(SubBlock, &Data);

		if(DataLength) {
			return Text(Data[0 .. DataLength-1]);
		} else {
			return Text.init;
		}
	}
	/**
	*/
	const Text fileInfo(TYPE Type, in LANGANDCODEPAGE LangCode) {
		return FileInfo(DEF.typeToDef(Type), LangCode);
	}

	/**
	History:
		$(HIST_VER 1.061,
			$(HIST_S ITextにあわせる。)
		)
	*/
	override const ConsText toText() {
		immutable LangAndCodes=langAndCode;
		
		Text[] infos;
		foreach(LangAndCode; LangAndCodes) {
			for(TYPE Type=TYPE.min; Type < TYPE.max; Type++) {
				immutable s=DEF.typeToDef(Type);
				const t=FileInfo(s, LangAndCode);
				if(t.length) {
					infos ~= Text("%s: %s", s, t);
				}
			}
		}
		return Text.join(infos, Text.newline);
	}
	/**
	*/
	override string toString() {
		return toText.toString;
	}
}
debug(exever) unittest {
	auto v=new Version(Text(`C:\Program Files\Internet Explorer\iexplore.exe`));
	//wl("%s", v.fileInfo(Version.TYPE.FILEVERSION, v.langAndCode[0]));
	//wl("%s",v.toText);
}

