﻿/**
ファイルとパス。

Phobosがchar[]大好きだったんで必要分だけ再構築。

std.pathとstd.fileの役割を統合。

History:
	$(HIST_VER 1.000,
		$(HIST_S NeGuiが依存している部分あったんでpackageをNeGuiに移動、全然GUIじゃない。)
	)
	$(HIST_VER 1.00β16,
		$(HIST_MSG std.fileをimportしないように変更。)
	)
*/
module etc.negui.file.file;

debug import etc.negui.nestdio: wl = writefln;
debug(file) void main() {}

import std.exception;
import std.traits;
import std.ctype;

import win32.windows;
import win32.shlwapi;
import win32.shlobj;

import etc.negui.system.base;
import etc.negui.system.type.record;
import etc.negui.system.utility.number;
import etc.negui.file.find;
import etc.negui.environ.security;
import etc.negui.containers.bit;
import etc.negui.com.imalloc;

//import hide_file=etc.negui.file._hide.file;
import hide_file=etc.negui.file.file_hide;

unittest {
	assert(!(PATH.MAX_NAME % 4));
	assert(!(PATH.MAX_PATH % 4));
}

///
abstract class NeGuiFileBaseException: NeGuiException {
	mixin MixInNeGuiException;
}
///
class NeGuiPathException: NeGuiFileBaseException {
	mixin MixInNeGuiException;
}
///
class NeGuiFileException: NeGuiFileBaseException {
	mixin MixInNeGuiException;
}
///
class DriveBaseException: NeGuiFileException {
	mixin MixInNeGuiException;
}
///
class DriveException: DriveBaseException {
	mixin MixInNeGuiException;
}
///
class DiskException: DriveBaseException {
	mixin MixInNeGuiException;
}
///
class NeGuiFolderException: NeGuiFileException {
	mixin MixInNeGuiException;
}


/**
ファイルパス。

std.pathと同じく文字列のみに作用。
パスっぽい操作でも実ファイルが必要なものはFILEで。

NeGuiの基準
$(LIST_UNORDERED
	$(LIST_ITEM フォルダやディレクトリの呼称はfolderで統一。)
	$(LIST_ITEM フォルダの最後は\)
)

History:
	$(HIST_VER 1.110,
		$(HIST_D modulePath, myPath, myFolder, currentFolder, toShortPath, toLongPathをFILEへ移動。)
	)
*/
static struct PATH {
	static:

	/**
	History:
		$(HIST_VER 1.00β14,
			$(HIST_S nemuxi.baseから引越し。)
		)
	*/
	enum {
		/// 最大ファイル名長
		MAX_NAME = 255 + 1,
		/// 最大パス長。
		MAX_PATH = 32767 + MAX_NAME + 1,
	}
	/// パス中でのフォルダ区切り文字。
	static immutable Text sep={data:cast(tchar[])[std.path.sep[0]]};
	/// パスとパスの区切り文字。
	static immutable Text pathsep={data:cast(tchar[])[std.path.pathsep[0]]};
	/// カレントフォルダ。
	static immutable Text curdir={data:cast(tchar[])[std.path.curdir[0]]};
	/// 親フォルダを表す文字列。
	static immutable Text pardir={data:cast(tchar[])[std.path.pardir[0], std.path.pardir[1]]};

	/**
	フォルダパス判定。

	NeGui基準で最後がPATH.sepをフォルダとする。

	Params:
		Path = 調べるパス。

	Returns:
		フォルダならtrue、違えばfalse。

	History:
		$(HIST_VER 1.130,
			$(HIST_CHG isFolderPath, isFolder)
		)
		$(HIST_VER 1.100,
			$(HIST_S PATH.sepの使用。)
		)
	*/
	bool isFolder(in Text Path) {
		//return Path.length && Path[Path.length-1] == '\\';
		return Path.length && Path[Path.length - 1] == sep[0];
	}
	debug(file) unittest {
		assert(isFolder(Text(`asd\asd\`w)));
		assert(!isFolder(Text(`asd\asd`c)));
	}
	alias isFolder isFolderPath;

	/**
	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	bool isUNC(in Text Path) {
		return cast(bool)PathIsUNC(Path.ptr);
	}
	bool isUNCServer(in Text Path) {
		return cast(bool)PathIsUNCServer(Path.ptr);
	}
	bool isUNCServerShare(in Text Path) {
		return cast(bool)PathIsUNCServerShare(Path.ptr);
	}
	
	
	/**
	パスをフォルダに。

	Params:
		Path = パス。

	Returns:
		変換後のパス。
		Pathが元からフォルダならPath自身を返す。

	History:
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
			$(HIST_S PATH.sepの使用。)
		)
	*/
	ref Text addFolderSep(in Text Path) {
		if(isFolderPath(Path)) {
			return cast(Text)Path;
		}
		//return Path ~ Text('\\');
		return Path ~ sep;
	}
	debug(file) unittest {
		assert(addFolderSep(Text(``)) == PATH.sep);
		assert(addFolderSep(Text(`asd\asd\`)) == Text(`asd\asd\`));
		assert(addFolderSep(Text(`asd\asd`)) == Text(`asd\asd\`));
	}
	
	/**
	フォルダから非フォルダパスへ。

	終端のフォルダ区切り文字を削除。

	Params:
		Path = パス。

	Returns:
		変換後のパス。
		Pathの終端がフォルダ区切り文字でない場合はPathを。

	History:
		$(HIST_VER 1.130,
			$(HIST_CHG 名前, delFolderSep, removeFolderSep)
		)
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
		)
	*/
	ref Text removeFolderSep(in Text Path) {
		if(auto len=Path.length) {
			while(isFolderPath(Path[0 .. len])) {
				if(len--) continue; else break;
			}
			return Path[0 .. len];
		} else {
			return Text.init;
		}
	}
	debug(file) unittest {
		assert(removeFolderSep(Text(`asd\asd\`)).data == `asd\asd`);
		assert(removeFolderSep(Text(`asd\asd`)).data == `asd\asd`);
		assert(removeFolderSep(Text(``)).data == ``);
		assert(removeFolderSep(Text(`a`)).data == `a`);
		assert(removeFolderSep(Text(`a\\\`)).data == `a`);
		assert(removeFolderSep(Text(`LICENSE`)).data == `LICENSE`);
	}
	/**
	delFolderSep置き換え。

	Deprecated:
		以後removeFolderSepを使用。
	*/
	alias removeFolderSep delFolderSep;
	
	/**
	パスの結合。

	Params:
		RootPath = 結合元となるパス。

		AddPath = RootPathに結合するパス。
	
	Returns:
		結合後のパス。
	
	History:
		$(HIST_VER 1.081,
			$(HIST_NEW)
		)
	*/
	Text join(in Text RootPath, in Text[] AddPath ... )
	out(r) {
		if(isFolderPath(AddPath[$-1])) {
			assert(isFolderPath(r));
		} else {
			assert(!isFolderPath(r));
		}
	}
	body {
		auto JoinPath=addFolderSep(RootPath);
		foreach(Path; AddPath) {
			JoinPath ~= addFolderSep(Path);
		}
		if(isFolderPath(AddPath[$-1])) {
			return JoinPath;
		} else {
			return removeFolderSep(JoinPath);
		}
	}
	debug(file) unittest {
		Text t1, t2, t3;
		t1 = Text("a");
		t2 = Text("b");
		t3 = Text("c");
		assert(join(t1, t2, t3).data == `a\b\c`);
		t3 ~= Text(`\`);
		assert(join(t1, t2, t3).data == `a\b\c\`);
	}

	/**
	拡張子取得。

	Params:
		Path = 拡張子を取得したいパス。

	Returns:
		拡張子。

	History:
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
			$(HIST_B 拡張子なしの場合に.なければエラー。)
		)
	*/
	ref Text getExtension(in Text Path) {
		auto pStr=Path.ptr(true);
		if(auto p=PathFindExtension(pStr)) {
			if(&pStr[Path.length] == p) {
				return Text.init;
			}
			return Text(cast(wchar*)(p+1));
		} else {
			return Text.init;
		}
	}
	debug(file) unittest {
		assert(getExtension(Text(``)).data == ``);
		
		assert(getExtension(Text(`123456`)).data == ``);
		assert(getExtension(Text(`123.56`)).data == `56`);
		assert(getExtension(Text(`.23.56`)).data == `56`);
		assert(getExtension(Text(`.23456`)).data == `23456`);
		assert(getExtension(Text(`abc\def`)) == Text(``));
		assert(getExtension(Text(`abc\.def`)) == Text(`def`));
		Text t;
		assert(getExtension(t) == Text(``));
	}
	
	/**
	拡張子の変更。

	Params:
		Path = パス名。

		Extension = 変更する拡張子。

	Returns:
		Pathに拡張子が無い場合は.Extensionが追加される。

	History:
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
		)
		$(HIST_VER 1.063,
			$(HIST_NEW)
		)
	*/
	ref Text changeExtension(in Text Path, in Text Extension) {
		if(auto ExtLength=getExtension(Path).length) {
			return Path[0 .. Path.length - ExtLength] ~ Extension;
		} else {
			return Path ~ Text('.') ~ Extension;
		}
	}
	debug(file) unittest {
		assert(changeExtension(Text(``), Text(``)) == Text(`.`),changeExtension(Text(``), Text(``)).toString);
		assert(changeExtension(Text(`abc.def`), Text(`ghi`)) == Text(`abc.ghi`));
		assert(changeExtension(Text(`abc`), Text(`def`)) == Text(`abc.def`));
		assert(changeExtension(Text(`.abc`), Text(`def`)) == Text(`.def`));
	}
	/**
	拡張子追加。

	changeExtensionと違い強制的に拡張子をつける。

	Params:
		Path = パス名。

		Extension = 結合する拡張子。

	Returns:
		Path ~ "." ~ Extensionを返す。

	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	ref Text addExtension(in Text Path, in Text Extension) {
		return Text.join([Path, Extension], Text('.'));
	}
	debug(file) unittest {
		assert(addExtension(Text("a"), Text("b")) == Text("a.b"));
	}

	
	/**
	ファイル名取得。

	Params:
		path = ファイル名を取得したいパス。

	Returns:
		拡張子付きのファイル名。
	
	History:
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
		)
	*/
	ref Text getFileName(in Text path) {
		return Text(PathFindFileName(path.ptr));
	}
	debug(file) unittest {
		assert(getFileName(Text(``)).data == ``);
		assert(getFileName(Text(`Z:\aaa.txt`)).data == `aaa.txt`);
		assert(getFileName(Text(`Z:\aaatxt`)).data == `aaatxt`);
		assert(getFileName(Text(`Z:\aaatxt\`)).data == `aaatxt\`);
		assert(getFileName(Text(`LICENSE`)).data == `LICENSE`);
	}
	/**
	ファイル名取得。

	Params:
		path = ファイル名を取得したいパス。

	Returns:
		拡張子無しのファイル名。

	History:
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
		)
		$(HIST_VER 1.00β16,
			$(HIST_F 拡張子のみ(.htaccess等の隠しファイル)をファイル名として扱うように変更。)
		)
		$(HIST_VER 1.00β11,
			$(HIST_B スライスで死ぬ対策。)
		)
	*/
	ref Text getName(in Text path) {
		auto name=removeFolderSep(getFileName(path));
		
		auto ext =getExtension(path);
		
		if(ext.length && name.length > ext.length+1) {
			return name[0..name.length - ext.length-1];
		} else {
			return name[];
		}
	}
	debug(file) unittest {
		assert(getName(Text(``)).data == ``);
		assert(getName(Text(`.`)).data == `.`);
		assert(getName(Text(`.aaa`)).data == `.aaa`);
		assert(getName(Text(`Z:\aaa.txt`)).data == `aaa`);
		assert(getName(Text(`Z:\aaatxt\`)).data == `aaatxt`);
		assert(getName(Text(`aaabbb`)).data == `aaabbb`);
		assert(getName(Text(`aaa.bbb`)).data == `aaa`);
		assert(getName(Text(`aaa.bbb.ccc`)).data == `aaa.bbb`);
	}
	/**
	親フォルダ取得。

	Params:
		Path = 親フォルダを取得したいパス。

	Returns:
		親フォルダ。

	Out:
		戻り値に長さがあれば最後は\で終わってる。

	History:
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
			$(HIST_S PATH.sepの使用。)
		)
		$(HIST_VER 1.100,
			$(HIST_S Phobosを使用しない様に変更。)
			$(HIST_CHG 名前, getOwnerFolder, ownerFolder)
		)
	*/
	ref Text ownerFolder(in Text Path) {
		if(Path.length) {
			auto folder=removeFolderSep(Path);
			for(auto i=folder.length-1; i != 0; i--) {
				if(folder[i] == sep[0]) {
					return folder[0 .. i+1];
				} else if(folder[i] == ':') {
					//return folder[0 .. i+1] ~ Text('\\');
					return addFolderSep(folder[0 .. i+1]);
				}
			}
		}

		return Text(cast(tstring)['.', sep[0]]);
	}
	debug(file) unittest {
		assert(ownerFolder(Text(``)).data == `.\`);
		assert(ownerFolder(Text(`123`)).data == `.\`);
		assert(ownerFolder(Text(`123\456`)).data == `123\`);
		assert(ownerFolder(Text(`123\456\`)).data == `123\`);
		assert(ownerFolder(Text(`123\456\789`)).data == `123\456\`);
	}

	/**
	ファイルパス比較。

	単純にファイル文字列の比較であって相対パス等は考慮しない。
	
	History:
		$(HIST_VER 1.00β11,
			$(HIST_NEW)
		)
	*/
	int cmp(in Text path1, in Text path2) {
		return path1.icmp(path2);
	}
	
	/**
	History:
		$(HIST_VER 1.120,
			$(HIST_NEW)
		)
	*/
	bool match(in Text Path, in Text Pattern) {
		return cast(bool)PathMatchSpec(Path.ptr, Pattern.ptr);
	}
	debug(file) unittest {
		assert(match(Text("a.a"), Text("*")));
		assert(match(Text("a.a.a"), Text("*.a.?")));
	}

	/**
	History:
		$(HIST_VER 1.120,
			$(HIST_NEW)
		)
	*/
	bool isContent(in Text Path, in Text ContentType) {
		return cast(bool)PathIsContentType(Path.ptr, ContentType.ptr);
	}
	debug(file) unittest {
		assert(isContent(Text("a.txt"), Text("text/plain")));
		assert(isContent(Text("a.css"), Text("text/css")));
		assert(!isContent(Text("a.js"), Text("text/js")));
		assert(!isContent(Text("a.js"), Text("text/ecmascript")));
	}

	///
	immutable dchar[] NO_FILE_CHARACTER = "\\/:*?\"<>|";

	/**
	ファイル名の使用判定。

	Params:
		c = ファイルの文字。

		FileName = ファイル名。

	Returns:
		有効無効の真偽値。
	
	See_Also:
		$(LIST_UNORDERED
			$(LIST_ITEM $(LINK http://support.microsoft.com/kb/405917/ja?FR=1&PA=1&SD=HSCH))
			$(LIST_ITEM $(LINK http://support.microsoft.com/kb/879124/ja))
		)

	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	bool isName(dchar c) {
		return !IsIn(c, cast(dchar[])NO_FILE_CHARACTER);
	}
	debug(file) unittest {
		assert(isName('a'));
		assert(!isName('\\'));
		assert(!isName('/'));
		assert(!isName(':'));
		assert(!isName('*'));
		assert(!isName('/'));
		assert(!isName('\"'));
		assert(!isName('<'));
		assert(!isName('>'));
		assert(!isName('|'));
	}
	/**
	ファイル名の使用判定。

	Params:
		FileName = ファイル名。

	Returns:
		有効無効の真偽値。

	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	bool isName(in Text FileName) {
		bool run=true;
		
		foreach(i, c; FileName.data) {
			if(!isName(c)) {
				return false;
			}

			if(run || i <= "con".length - 1) {
				auto Head=Text.iwrap(cast(tstring)FileName.data[0 .. 3]).toUpper.data.idup;

				if(IsIn(Head, [cast(tstring)"PRN", "AUX", "NUL", "CON"])) {
					if(FileName.data.length == "con".length) {
						return false;
					}
					assert(FileName.data.length > "con".length);
					return FileName.data[3] != '.';
				}
				
				if(FileName.data.length == "comN".length && IsIn(Head, [cast(tstring)"COM", "LPT"])) {
					return !cast(bool)isdigit(FileName.data[3]);
				}
				
				if(i >= "comN".length - 1) {
					run = false;
				}
			}
		}
		
		return true;
	}
	debug(file) unittest {
		assert(isName(Text("aaa")));
		assert(isName(Text("あいうえお")));
		assert(!isName(Text("あいうえお|")));
		
		assert(!isName(Text("a?a")));

		
		assert(!isName(Text("con")));
		assert(!isName(Text("CON")));
		assert(!isName(Text("CON.")));
		assert(!isName(Text("CON.1")));
		assert(!isName(Text("CON.CON")));
		assert(isName(Text("CONCON")));
		assert(isName(Text("CONSOLE")));
		
		assert(!isName(Text("com1")));
		assert(!isName(Text("COM9")));
		assert(isName(Text("COM")));
		assert(isName(Text("COMA")));
		assert(isName(Text("COM10")));
	}

	/**
	$(STRONG .)や$(STRONG ..)等を含むパス名の変換。
	
	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	ref Text toCanonicalize(in Text SourcePath) {
		tchar[MAX_PATH] Path;
		enforce(PathCanonicalize(Path.ptr, SourcePath.ptr), new NeGuiPathException(ERR.toText));
		return Text(Path.ptr);
	}
	debug(file) unittest {
		// http://msdn.microsoft.com/en-us/library/bb773569%28VS.85%29.aspx
		assert(toCanonicalize(Text(`A:\name_1\.\name_2\..\name_3`)) == Text(`A:\name_1\name_3`));
		assert(toCanonicalize(Text(`A:\name_1\..\name_2\.\name_3`)) == Text(`A:\name_2\name_3`));
		assert(toCanonicalize(Text(`A:\name_1\name_2\.\name_3\..\name_4`)) == Text(`A:\name_1\name_2\name_4`));
		assert(toCanonicalize(Text(`A:\name_1\.\name_2\.\name_3\..\name_4\..`)) == Text(`A:\name_1\name_2`));
		assert(toCanonicalize(Text(`C:\..`)) == Text(`C:\`));
	}


	/**
	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	ref Text toRelative(in Text BasePath, in Text TartgetPath) {
		auto BasePathAttribute  =isFolderPath(BasePath) ? FILE_ATTRIBUTE_DIRECTORY: FILE_ATTRIBUTE_NORMAL;
		auto TargetPathAttribute=isFolderPath(TartgetPath) ? FILE_ATTRIBUTE_DIRECTORY: FILE_ATTRIBUTE_NORMAL;
		
		tchar[MAX_PATH] Path;
		enforce(PathRelativePathTo(Path.ptr, BasePath.ptr, BasePathAttribute, TartgetPath.ptr, TargetPathAttribute), new NeGuiPathException(ERR.toText));
		return Text(Path.ptr);
	}
	/**
	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	ref Text toAbsolute(in Text BasePath, in Text RelativePath) {
		return toCanonicalize(join(BasePath, RelativePath));
	}
	debug(file) unittest {
		assert(toAbsolute(Text(`A:\name_1\`), Text(`.\name_2\..\name_3`)) == Text(`A:\name_1\name_3`));
		assert(toAbsolute(Text(`A:\name_1\`), Text(`..\name_3`)) == Text(`A:\name_3`));
	}
}




/**
ファイル。

色々あって実装してみたものの使う機会あんましない。
std.streamかstd.stdio.File使うべきだよ。

Bugs:
	試験してない。

History:
	$(HIST_VER 1.100,
		$(HIST_NEW)
	)
*/
class NeFile: HandleRaii {
	alias Handle hFile;
	mixin KillResource!("ファイル", hFile, CloseHandle, 0);

	protected bool IsSeek;
	mixin(SMixInClassGet!(typeof(IsSeek))("isSeek", q{IsSeek}, false));

	///
	enum ACCESS_MODE {
		NONE  = 0,             /// オブジェクトへのデバイス問い合わせアクセスを指定します。アプリケーションは、デバイスにアクセスすることなく、デバイス属性の問い合わせを行えます。
		READ  = GENERIC_READ,  /// オブジェクトへの読み取りアクセスを指定します。データの読み取りとファイルポインタの移動ができます。読み書きアクセスを行うには、GENERIC_WRITE と共に指定します。
		WRITE = GENERIC_WRITE, /// オブジェクトへの書き込みアクセスを指定します。データの書き込みとファイルポインタの移動ができます。読み書きアクセスを行うには、GENERIC_READ と共に指定します。
		ALL   = READ | WRITE,
	}

	///
	enum SHARE {
		NONE   = 0,
		DELETE = FILE_SHARE_DELETE, /// Windows NT/2000：これ以降、オブジェクトを開こうとしたときに削除アクセスを要求した場合のみ、その操作を許可します。
		READ   = FILE_SHARE_READ,   /// これ以降、オブジェクトを開こうとしたときに読み取りアクセスを要求した場合のみ、その操作を許可します。
		WRITE  = FILE_SHARE_WRITE,  /// これ以降、オブジェクトを開こうとしたときに書き込みアクセスを要求した場合のみ、その操作を許可します。
		ALL    = DELETE | READ | WRITE,
	}

	///
	enum DISPOSITION {
		NONE              = 0,
		CREATE_NEW        = CREATE_NEW        , /// 新しいファイルを作成します。指定したファイルが既に存在している場合、この関数は失敗します。
		CREATE_ALWAYS     = CREATE_ALWAYS     , /// 新しいファイルを作成します。指定したファイルが既に存在している場合、そのファイルを上書きし、既存の属性を消去します。
		OPEN_EXISTING     = OPEN_EXISTING     , /// ファイルを開きます。指定したファイルが存在していない場合、この関数は失敗します。デバイスに対して CreateFile 関数を呼び出す際に OPEN_EXISTING フラグを使うべき状況については、この関数の「解説」を参照してください。
		OPEN_ALWAYS       = OPEN_ALWAYS       , /// ファイルが存在している場合、そのファイルを開きます。指定したファイルが存在していない場合、この関数は dwCreationDisposition パラメータで CREATE_NEW が指定されていたと仮定して新しいファイルを作成します。
		TRUNCATE_EXISTING = TRUNCATE_EXISTING , /// ファイルを開き、ファイルのサイズを 0 バイトにします。呼び出し側プロセスは、dwDesiredAccess パラメータで、少なくとも GENERIC_WRITE アクセス権を指定しなければなりません。指定したファイルが存在していない場合、この関数は失敗します。		
	}
	///
	enum FLAG {
		NONE               = 0,
		WRITE_THROUGH      = FILE_FLAG_WRITE_THROUGH,      /// キャッシュに書き込まれたデータを直接ディスクに書き込むようシステムに指示します。システムは依然としてキャッシュ書き込み操作を行えますが、ディスクへの遅延書き込みは無効になります。
		OVERLAPPED         = FILE_FLAG_OVERLAPPED,         /// オブジェクトを初期化するようシステムに指示します。その結果、非常に長い時間のかかる処理に対して ERROR_IO_PENDING が返るようになります。この処理が終了すると、指定されたイベントはシグナル状態に設定されます。このフラグを指定したときは、ファイルの読み書きを行う関数で 1 個の 構造体を指定しなければなりません。つまり、FILE_FLAG_OVERLAPPED を指定した場合、アプリケーションはオーバーラップ読み書きを行わなければなりません。FILE_FLAG_OVERLAPPED を指定した場合、システムはファイルポインタを管理しません。ファイルの読み書きを行う関数に対し、1 個の OVERLAPPED 構造体を指している lpOverlapped パラメータの一部として、ファイルの位置を渡さなければなりません。また、このフラグを指定した場合、1 つのハンドルに対して複数の操作（たとえば、読み取りと書き込みそれぞれの操作）を同時に行うことができます。
		NO_BUFFERING       = FILE_FLAG_NO_BUFFERING,       /// バッファやキャッシュを使わずにファイルを開くようシステムに指示します。FILE_FLAG_OVERLAPPED フラグと共に指定すると、非同期操作のパフォーマンスが向上します。I/O 操作を行う際に、メモリマネージャの同期処理に依存する必要がなくなるからです。しかし、データがキャッシュ内に存在しないので、一部の I/O 操作には多くの時間がかかります。このフラグを指定した場合、アプリケーションは次の各条件を満たさなければなりません。$(LIST_UNORDERED $(LIST_ITEM ファイルアクセスの開始オフセットを、ボリュームのセクタサイズの整数倍にしなければなりません。)$(LIST_ITEM ファイルアクセスのバイト数を、ボリュームのセクタサイズの整数倍にしなければなりません。たとえば、セクタサイズが 512 バイトの場合、512 バイト、1,024 バイト、2,048 バイトの読み書きはできますが、335 バイト、981 バイト、7,171 バイトなどの読み書きはできません。)$(LIST_ITEM 読み書き操作用のバッファのアドレス（メモリ内のアドレス）を、ボリュームのセクタサイズの整数倍に整列（セクタ整列）させなければなりません。ディスクによっては、この要件を満たせないこともあります。))3 番目の条件を満たす、つまりバッファをボリュームのセクタサイズの整数倍の位置に整列させる 1 つの方法は、 関数を使ってバッファを割り当てることです。VirtualAlloc 関数が割り当てるメモリは、オペレーティングシステムのメモリページサイズの整数倍の位置に整列されています。メモリページとボリュームのセクタサイズは両方とも 2 のべき乗なので、このメモリは、ボリュームのセクタサイズの整数倍のアドレスにも整列されます。ボリュームのセクタサイズを求めるには、GetDiskFreeSpace 関数を使います。また、ページサイズを調べるには、 関数を使います。
		RANDOM_ACCESS      = FILE_FLAG_RANDOM_ACCESS,      /// このファイルをランダムアクセスすることをシステムに指示します。システムは、この指定をファイルのキャッシングを最適化するためのヒントとして使います。
		SEQUENTIAL_SCAN    = FILE_FLAG_SEQUENTIAL_SCAN,    /// ファイルを最初から最後までシーケンシャルに（連続）アクセスすることをシステムに指示します。システムは、この指定をファイルのキャッシングを最適化するためのヒントとして使います。この場合、アプリケーションはランダムアクセスを行う目的でファイルポインタを移動することはできますが、最適なキャッシングが達成できなくなる可能性があります。しかし、ランダムアクセスを正しく行えることは保証されています。このフラグを指定すると、大きなファイルをシーケンシャルアクセスするアプリケーションのパフォーマンスが向上します。基本的にシーケンシャルアクセスを行い、ときどき小規模な範囲をスキップするアプリケーションでは、特に大きな向上が見られることもあります。
		DELETE_ON_CLOSE    = FILE_FLAG_DELETE_ON_CLOSE,    /// FILE_FLAG_DELETE_ON_CLOSE を指定したハンドルを除き、このファイルに関連付けられているすべてのハンドルが閉じた段階で、そのファイルを即座に削除するようシステムに指示します。それ以降、FILE_SHARE_DELETE を指定しない限り、このファイルを開くよう要求しても失敗します。
		BACKUP_SEMANTICS   = FILE_FLAG_BACKUP_SEMANTICS,   /// Windows NT/2000：バックアップまたは復元操作の目的で、ファイルを開くまたは作成するようシステムに指示します。呼び出し側プロセスがファイルセキュリティに関して十分な権限を備えていることを前提として、システムは、そのプロセスがファイルセキュリティのチェックを無効にできることを保証します。これに関連する権限は、 SE_BACKUP_NAME と SE_RESTORE_NAME です。ディレクトリのハンドルを取得する目的で、このフラグを指定することもできます。特定の Win32 関数で、ファイルハンドルの代わりにディレクトリハンドルを指定することもできます。
		POSIX_SEMANTICS    = FILE_FLAG_POSIX_SEMANTICS,    /// POSIX（UNIX ベースのポータブル OS）の規則に従ってファイルにアクセスするようシステムに指示します。この規則には、ファイル名が同じで大文字と小文字だけが異なる複数のファイルを認めるファイルシステムに対し、それらのファイルへのアクセスを認めることが含まれます。このフラグを指定する場合、注意が必要です。MS-DOS または 16 ビット版 Windows からは、このフラグを指定して作成したファイルにアクセスできないからです。
		OPEN_REPARSE_POINT = FILE_FLAG_OPEN_REPARSE_POINT, /// このフラグを指定すると、NTFS の再解析ポイントで再解析を行うことを禁止します。このフラグを指定してファイルを開くと、再解析ポイントを制御するフィルタが動作しているかどうかにかかわりなく、ファイルハンドルが返ります。このフラグと共に CREATE_ALWAYS フラグを指定することはできません。
		OPEN_NO_RECALL     = FILE_FLAG_OPEN_NO_RECALL,     /// ファイルのデータを要求すると共に、そのデータを継続的にリモート記憶域に配置するべきことをシステムに指示します。そのファイルをローカル記憶域へ復帰させるべきではありません。このフラグは、リモート記憶域システム、言い換えると階層記憶管理システムで使うことを意図しています。
	}
	///
	override this(HANDLE Handle, bool Suicide, bool IsSeek) {
		this.IsSeek = IsSeek;
		super(Handle, Suicide);
	}
	///
	protected this(in Text Path, ACCESS_MODE AccessMode, SHARE Shere, in SECURITY* Security, DISPOSITION Disposition, FILE.ATTRIBUTE Attribute, FLAG Flags, NeFile Template, bool IsSeek) {
		ERR.initialize;
		
		auto hHandle = CreateFile(
			Path.ptr,
			AccessMode,
			Shere,
			Security ? Security.ptr: null,
			Disposition,
			Attribute | Flags,
			Template ? Template(): null
		);
		enforce(hFile != INVALID_HANDLE_VALUE, new NeGuiFileException(Text("%s, %s", ERR.toText, Path)));
		this(hHandle, true, IsSeek);
	}
	///
	this(in Text Path, ACCESS_MODE AccessMode, SHARE Shere, const ref SECURITY Security, DISPOSITION Disposition, FILE.ATTRIBUTE Attribute, FLAG Flags, NeFile Template, bool IsSeek) {
		this(Path, AccessMode, Shere, &Security, Disposition, Attribute, Flags, Template, IsSeek);
	}
	///
	this(in Text Path, ACCESS_MODE AccessMode, SHARE Shere, DISPOSITION Disposition, FILE.ATTRIBUTE Attribute, FLAG Flags, bool IsSeek) {
		this(Path, AccessMode, Shere, null, Disposition, Attribute, Flags, null, IsSeek);
	}
	private static NeFile rw(in Text Path, bool Write=false) {
		return new NeFile(
			Path,
			Write ? ACCESS_MODE.READ | ACCESS_MODE.WRITE: ACCESS_MODE.READ,
			SHARE.ALL,
			Write ? DISPOSITION.OPEN_ALWAYS: DISPOSITION.OPEN_EXISTING,
			FILE.ATTRIBUTE.NORMAL,
			FLAG.NONE,
			true
		);
	}
	///
	static NeFile reader(in Text Path) {
		return rw(Path, false);
	}
	///
	static NeFile writer(in Text Path) {
		return rw(Path, true);
	}

	///
	enum TYPE {
		UNKNOWN = FILE_TYPE_UNKNOWN, /// 指定されたファイルの種類は不明です。
		DISK    = FILE_TYPE_DISK,    /// 指定されたファイルは、ディスクファイルです。
		CHAR    = FILE_TYPE_CHAR,    /// 指定されたファイルは、LPT デバイスやコンソールのような文字ファイルです。
		PIPE    = FILE_TYPE_PIPE,    /// 指定されたファイルは、名前付きまたは名前なしパイプです。
		//REMOTE  = FILE_TYPE_REMOTE,  ///
	}
	///
	const TYPE type() {
		return cast(TYPE)GetFileType(hFile);
	}


	
	/**
	この関数の動作を修飾するフラグを指定します。このパラメータで、次の値のいずれかまたは任意の組み合わせを指定します。
	*/
	enum LOCK {
		FAIL_IMMEDIATELY = LOCKFILE_FAIL_IMMEDIATELY, /// 要求したロックを獲得できなかった場合、この関数はすぐに制御を返します。この値を指定しなかった場合は、ロックを獲得できるまで待ち続けます。
		EXCLUSIVE_LOCK   = LOCKFILE_EXCLUSIVE_LOCK,   /// この関数は、排他ロックを要求します。この値を指定しなかった場合は、共有ロックを要求します。
	}
	///
	bool lock(const ref LARGE Offset, const ref LARGE Size) {
		return cast(bool)LockFile(hFile, Offset.low, Offset.high, Size.low, Size.high);
	}
	///
	bool lock(LOCK Lock, const ref LARGE Size, ref OVERLAPPEDINFO OverlappedInfo) {
		return cast(bool)LockFileEx(hFile, Lock, 0, Size.low, Size.high, OverlappedInfo.ptr);
	}
	///
	bool unLock(const ref LARGE Offset, const ref LARGE Size) {
		return cast(bool)UnlockFile(hFile, Offset.low, Offset.high, Size.low, Size.high);
	}
	///
	bool unLock(const ref LARGE Size, ref OVERLAPPEDINFO OverlappedInfo) {
		return cast(bool)UnlockFileEx(hFile, 0, Size.low, Size.high, OverlappedInfo.ptr);
	}

	///
	enum SEEK {
		TOP = FILE_BEGIN,   /// 0、またはファイルの先頭が開始点になります。このフラグを指定すると、liDistanceToMove パラメータは符号のない値として解釈されます。
		NOW = FILE_CURRENT, /// 現在のファイルポインタの値が開始点になります。
		END = FILE_END,     /// 現在ファイルの終端の位置が開始点になります。
	}
	/**
	現在位置の変更。

	Returns:
		変更後の位置、
	*/
	const LARGE seekTo(const ref LARGE Position, SEEK Seek) {
		LARGE Result;
		enforce(SetFilePointerEx(hFile, Position.LargeInteger, Result.ptr, Seek), new NeGuiFileException(Text("%s, IsSeek -> %s", ERR.toText, IsSeek)));
		return Result;
	}
	const LARGE seekTo(long Position, SEEK Seek) {
		LARGE Large;
		Large.total = Position;
		return seekTo(Large, Seek);
	}
	

	/// 現在位置
	@property const LARGE position() {
		LARGE Large;
		Large.total = 0;
		return seekTo(Large, SEEK.NOW);
	}
	/// ditto
	@property const LARGE position(LARGE Large) {
		return seekTo(Large, SEEK.NOW);
	}

	/// 現在位置を先頭へ。
	bool seekToTop() {
		LARGE Large;
		return collectException(seekTo(Large, SEEK.TOP)) is null;
	}
	/// 現在位置を終端へ。
	bool seekToEnd() {
		LARGE Large;
		return collectException(seekTo(Large, SEEK.END)) is null;
	}
	
	///
	bool eof() {
		auto NowPosition=position;
		seekToEnd();
		auto LastPosition=position;
		scope(exit) position = NowPosition;
		
		return NowPosition == LastPosition;
	}
	///
	bool setEof() {
		return cast(bool)SetEndOfFile(hFile);
	}
	
	///
	final bool flush() {
		return cast(bool)FlushFileBuffers(hFile);
	}
	///
	final bool cancel() {
		return cast(bool)CancelIo(hFile);
	}

	protected {
		/**
		*/
		size_t Read(void* Buffer, size_t BufferSize) {
			size_t Size;

			enforce(ReadFile(hFile, Buffer, BufferSize, &Size, null), new NeGuiFileException(ERR.toText));
			
			return Size;
		}
		/**
		*/
		size_t Write(in void* Buffer, size_t BufferSize)
		in {
			assert(BufferSize, "use timeStamp()");
		}
		body {
			size_t Size;

			enforce(WriteFile(hFile, cast(void*)Buffer, BufferSize, &Size, null), new NeGuiFileException(ERR.toText));
			
			return Size;
		}

		void SeekExec(void delegate() SeekDg, void delegate() NonSeekDg=null) {
			if(IsSeek) {
				SeekDg();
			} else if(NonSeekDg) {
				NonSeekDg();
			}
		}
	}
	/**
	読み込み。

	Params:
		Buffer = 読み込む領域。

	Returns:
		正常に読み込めた場合はtrue、失敗した場合はfalseを返す。
		仮にshortを読み込もうとしてbyteしか読めなかった場合はシークを呼び出し前の位置に直してfalseを返す。
	*/
	bool get(T)(out T Buffer)
	if(!isArray!(T))
	{
		LARGE OldSeek;
		SeekExec({OldSeek=position;});
		auto Size=Read(&Buffer, T.sizeof);
		if(Size == T.sizeof) {
			return true;
		} else {
			SeekExec({seekTo(OldSeek, SEEK.TOP);});
			return false;
		}
	}
	/**
	読み込み。

	Params:
		Buffer = 読み込む領域。

	Returns:
		Bufferの有効なインデックスを返す。
		仮にint[4](16-Byte)を読み込もうとして10-Byteしか読めなかった場合はシークを8-Byte分読み込んだ部分に戻して2(読み込めた8-Byte分)を返す。
	*/
	size_t read(T)(ref T[] Buffer)
	if(isArray!(T[]))
	{
		LARGE OldSeek;
		SeekExec({OldSeek=position;});
		auto Size=Read(Buffer.ptr, T.sizeof * Buffer.length);
		auto SafeIndex=Size / T.sizeof;
		if(SafeIndex != Buffer.length) {
			SeekExec({seekTo(OldSeek.total + T.sizeof * SafeIndex, SEEK.TOP);});
		}
		return SafeIndex;
	}

	/**
	書き込み。

	Params:
		Buffer = 書き込む値。

	Returns:
		正常に書き込めた場合はtrue、失敗した場合はfalseを返す。
		仮にshortを書き込もうとしてbyteしか書けなかった場合はシークを呼び出し前の位置に直してfalseを返す。
	*/
	bool put(T)(T Buffer)
	if(!isArray!(T))
	{
		LARGE OldSeek;
		SeekExec({OldSeek=position;});
		auto Size=Write(&Buffer, T.sizeof);
		if(Size == T.sizeof) {
			return true;
		} else {
			SeekExec({seekTo(OldSeek, SEEK.TOP);});
			return false;
		}
	}
	/**
	書き込み。

	Params:
		Buffer = 書き込む値。

	Returns:
		書き込まれたBufferの有効なインデックスを返す。
		仮にint[4](16-Byte)を書き込もうとして10-Byteしか書けなかった場合はシークを8-Byte分書き込んだ部分に戻して2(書き込めた8-Byte分)を返す。
	*/
	size_t write(T)(in T[] Buffer)
	if(isArray!(T[]))
	{
		LARGE OldSeek;
		SeekExec({OldSeek=position;});
		auto Size=Write(Buffer.ptr, T.sizeof * Buffer.length);
		auto SafeIndex=Size / T.sizeof;
		if(SafeIndex != Buffer.length) {
			SeekExec({seekTo(OldSeek + T.sizeof * SafeIndex, SEEK.TOP);});
		}
		return SafeIndex;
	}
}


///
static struct FILE {
	static:

	/**
	モジュールのパスを取得。
	
	History:
		$(HIST_VER 1.110,
			$(HIST_CHG 構造体, PATH, FILE)
			$(HIST_CHG 例外, NeGuiPathException, NeGuiFileException)
		)
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
		)
		$(HIST_VER 1.00β17,
			$(HIST_NEW)
		)
	*/
	ref Text modulePath(HINSTANCE hInstance, size_t Length=PATH.MAX_PATH) {
		auto text=new wchar[Length];

		/+
		if(!GetModuleFileName(hInstance, text.ptr, text.length)) {
			throw new NeGuiPathException(ERR.toText);
		}
		+/
		enforce(GetModuleFileName(hInstance, text.ptr, text.length), new NeGuiFileException(ERR.toText));

		return Text(text.ptr);
	}

	/**
	自身のアドレス取得。
	
	History:
		$(HIST_VER 1.110,
			$(HIST_CHG 構造体, PATH, FILE)
			$(HIST_CHG 例外, NeGuiPathException, NeGuiFileException)
		)
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
		)
		$(HIST_VER 1.00β14,
			$(HIST_MSG MyPathの引数追加に合わせてデフォルト引数の追加。)
		)
		$(HIST_VER 1.00β11,
			$(HIST_NEW)
		)
	*/
	ref Text myPath(size_t Length=PATH.MAX_PATH) {
		try {
			return modulePath(GetModuleHandle(null), Length);
		} catch(NeGuiFileException e) {
			throw new NeGuiPathException(Text("自分の居場所さえ分からない"), e);
		}
	}
	/**
	自身の親フォルダ。
	
	History:
		$(HIST_VER 1.110,
			$(HIST_CHG 構造体, PATH, FILE)
			$(HIST_CHG 例外, NeGuiPathException, NeGuiFileException)
		)
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
		)
		$(HIST_VER 1.00β14,
			$(HIST_MSG MyPathの引数追加に合わせてデフォルト引数の追加。)
		)
		$(HIST_VER 1.00β11,
			$(HIST_MSG nemuxi.baseから移動。)
		)
	*/
	ref Text myFolder(size_t Length=PATH.MAX_PATH) {
		return PATH.ownerFolder(myPath(Length));
	}

	/**
	カレントフォルダ取得。
	
	History:
		$(HIST_VER 1.110,
			$(HIST_CHG 構造体, PATH, FILE)
			$(HIST_CHG 例外, NeGuiPathException, NeGuiFileException)
		)
		$(HIST_VER 1.100,
			$(HIST_S 戻り値をrefへ。)
		)
		$(HIST_VER 1.100,
			$(HIST_NEW)
		)
	*/
	ref Text currentFolder() {
		immutable BUFFER=GetCurrentDirectory(0, null);
		auto path=new wchar[BUFFER];

		auto CmpLength=GetCurrentDirectory(path.length, path.ptr);
		/+
		if(!CmpLength) {
			new NeGuiPathException(Text("カレントディレクトリ取得失敗"));
		}
		+/
		enforce(CmpLength, new NeGuiFileException(ERR.toText));
		enforce(CmpLength == path.length-1, new NeGuiFileException(Text("カレントディレクトリ取得後に変わった？")));

		//Text text;
		auto text=new Text;
		text.data = path[0 .. $];
		//text.data[$-1] = '\\';
		text.data[$-1] = PATH.sep[0];//'\\';
		
		return *text;
	}
	/**
	History:
		$(HIST_VER 1.110,
			$(HIST_CHG 構造体, PATH, FILE)
			$(HIST_CHG 例外, NeGuiPathException, NeGuiFileException)
		)
	*/
	bool currentFolder(in Text Path, bool ThrowException=false) {
		auto ret=cast(bool)SetCurrentDirectory(Path.ptr);

		if(!ret) {
			enforce(ThrowException, new NeGuiFileException(ERR.toText));
		}
		
		return ret;
	}


	private {
		alias extern(Windows) DWORD function(LPCTSTR, LPTSTR, DWORD) FuncType;
		ref Text ConvertShortLongName(in Text Path, FuncType Func) {
			ERR.initialize;
			const pPath=Path.ptr;
			immutable Length=Func(pPath, null, 0);

			enforce(Length, new NeGuiFileException(Text("%s, Path -> %s", ERR.toText, Path)));
			auto pResult=new wchar[Length];
			enforce(Length-1==Func(pPath, pResult.ptr, Length), new NeGuiFileException(ERR.toText));

			return Text.wrap(pResult[0..Length-1]);
		}
	}

	/**
	短いファイル名。

	History:
		$(HIST_VER 1.110,
			$(HIST_CHG 構造体, PATH, FILE)
			$(HIST_CHG 例外, NeGuiPathException, NeGuiFileException)
		)
	*/
	ref Text toShortPath(in Text Path) {
		return ConvertShortLongName(Path, cast(FuncType)&GetShortPathName);
	}
	/**
	長いファイル名。

	History:
		$(HIST_VER 1.110,
			$(HIST_CHG 構造体, PATH, FILE)
			$(HIST_CHG 例外, NeGuiPathException, NeGuiFileException)
		)
	*/
	ref Text toLongPath(in Text Path) {
		return ConvertShortLongName(Path, cast(FuncType)&GetLongPathName);
	}
	
	///
	enum ATTRIBUTE {
		NONE                = 0xFFFFFFFF,
		ARCHIVE             = FILE_ATTRIBUTE_ARCHIVE,             /// アーカイブファイルまたはアーカイブディレクトリです。アプリケーションはこの属性を、ファイルのバックアップや削除のためのマークとして使います。
		COMPRESSED          = FILE_ATTRIBUTE_COMPRESSED,          /// 指定されたファイルまたはディレクトリは圧縮されています。ファイルの場合、ファイル内の全データが圧縮されていることを意味します。ディレクトリの場合、そのディレクトリ内に新しく作成されるファイルまたはサブディレクトリが、既定で圧縮状態になることを意味します。
		DEVICE              = FILE_ATTRIBUTE_DEVICE,              /// 予約済み。使わないでください。
		FOLDER              = FILE_ATTRIBUTE_DIRECTORY,           /// 指定されたハンドルは、ディレクトリに関連しています。
		ENCRYPTED           = FILE_ATTRIBUTE_ENCRYPTED,           /// 指定されたファイルまたはディレクトリは暗号化されています。ファイルの場合、ファイル内の全データストリームが暗号化されていることを意味します。ディレクトリの場合、そのディレクトリ内に新しく作成されるファイルまたはサブディレクトリが、既定で暗号化状態になることを意味します。
		HIDDEN              = FILE_ATTRIBUTE_HIDDEN,              /// 隠しファイルまたは隠しディレクトリです。通常のディレクトリリスティングでは表示されません。
		NORMAL              = FILE_ATTRIBUTE_NORMAL,              /// 指定されたファイルまたはディレクトリには、特に属性はありません。単独で返った場合にのみ、この属性は有効です。
		NOT_CONTENT_INDEXED = FILE_ATTRIBUTE_NOT_CONTENT_INDEXED, /// Windows 2000：このファイルは、「インデックスサービス」の対象になっていません。
		OFFLINE             = FILE_ATTRIBUTE_OFFLINE,             /// Windows 2000：このファイルのデータは、すぐには利用できません。この属性は、ファイルのデータがオフラインの記憶装置へ物理的に移動されたことを示します。この属性は、 Windows 2000 の階層記憶管理ソフトウェアである「リモート記憶域」が利用するものです。アプリケーションは、任意にこの属性を変更するべきではありません。
		READONLY            = FILE_ATTRIBUTE_READONLY,            /// このファイルまたはディレクトリは読み取り専用です。アプリケーションはこのファイルの読み取りを行えますが、書き込みや削除はできません。ディレクトリの場合、アプリケーションは削除を行えません。
		REPARSE_POINT       = FILE_ATTRIBUTE_REPARSE_POINT,       /// このファイルには、再解析ポイントが関連付けられています。
		SPARSE_FILE         = FILE_ATTRIBUTE_SPARSE_FILE,         /// このファイルは、スパースファイル（疎なファイル、未使用の領域が多い、または同じ値が長く続くファイル）です。
		SYSTEM              = FILE_ATTRIBUTE_SYSTEM,              /// このファイルまたはディレクトリは、オペレーティングシステムの一部、またはオペレーティングシステム専用です。
		TEMPORARY           = FILE_ATTRIBUTE_TEMPORARY,           /// このファイルは、一時ファイルとして使われています。ファイルシステムは、データをハードディスクのような大容量記憶装置へ書き込む代わりに、高速なアクセスが行えるよう、すべてのデータをメモリ内に維持することを試みます。アプリケーションは、必要がなくなった段階で一時ファイルをすぐに削除するべきです。
	}
	/**
	*/
	ATTRIBUTE attributes(in Text path) {
		return cast(ATTRIBUTE)GetFileAttributes(path.ptr);
	}
	/// ditto
	bool attributes(in Text path, ATTRIBUTE Attribute) {
		return cast(bool)SetFileAttributes(path.ptr, Attribute);
	}
	/**
	指定アドレスは存在するか。

	Params:
		path = 調べたいアドレス。

	Returns:
		存在すればtrue、しなければfalse。
	*/
	bool isExistence(in Text path) {
		return attributes(path) != ATTRIBUTE.NONE;
	}
	/**
	指定アドレスはフォルダか。

	Params:
		Path = 調べたいアドレス。

	Returns:
		フォルダならtrue。

	Throws:
		アドレス自体が存在しなければNeGuiFileException。

	History:
		$(HIST_VER 1.130,
			$(HIST_P 例外時のメッセージ変更。)
		)
	*/
	bool isFolder(in Text Path) {
		enforce(isExistence(Path), new NeGuiFileException(Text("not found %s", Path)));
		return (attributes(Path) & ATTRIBUTE.FOLDER) != 0;
	}
	/**
	指定アドレスはファイルか。

	Params:
		path = 調べたいアドレス。

	Returns:
		ファイルならtrue。

	Throws:
		アドレス自体が存在しなければNeGuiFileException。
	*/
	bool isFile(in Text path) {
		return !isFolder(path);
	}

	
	/**
	フォルダ作成。

	Params:
		path = フォルダ

		Recurse = フォルダを再帰作成するか。

	Returns:
		成功したか否か。
	*/
	bool makeFolder(in Text path, bool Recurse=true) {
		if(Recurse) {
			const left=PATH.ownerFolder(path);
			isExistence(left) || makeFolder(left, true);
			return makeFolder(path, false);
		} else {
			return cast(bool)CreateDirectory(PATH.removeFolderSep(path).ptr, null);
		}
	}
	/**
	ファイル削除。

	History:
		$(HIST_VER 1.130,
			$(HIST_CHG 名前, delFile, removeFile)
		)
	*/
	bool removeFile(in Text path) {
		return cast(bool)DeleteFile(path.ptr);
	}
	/**
	delFile置き換え。

	Deprecated:
		以後removeFileを使用。
	*/
	alias removeFile delFile;
	
	/**

	Throws:
		Recurseがtrueでファイル列挙に失敗すればNeGuiExceptionを投げる
	
	History:
		$(HIST_VER 1.130,
			$(HIST_CHG 名前, delFolder, removeFolder)
		)
		$(HIST_VER 1.100,
			$(HIST_F ファイルの有無による処理。)
		)
	*/
	bool removeFolder(in Text path, bool Recurse=false) {
		if(Recurse) {
			try {
				enforce(isFolder(path), new NeGuiFileException(Text("%s is not found.", path)));
			} catch(Exception e) {
				throw new NeGuiFileException(Text("path -> %s", path), e);
			}
			
			FILEITEM FileItem=void;
			scope finder=new Finder(PATH.join(path, Text('*')));
			while(finder.find(FileItem)) if(FileItem.isFolderItems) {
				const FileName=PATH.join(path, FileItem.name);
				if(FileItem.isFolder) {
					removeFolder(FileName, true);
				} else {
					removeFile(FileName);
				}
			}
		}
		
		return cast(bool)RemoveDirectory(path.ptr);
	}
	/**
	delFolder置き換え。

	Deprecated:
		以後removeFolderを使用。
	*/
	alias removeFolder delFolder;

	private enum {
		MOVEFILE_REPLACE_EXISTING       = 0x00000001,
		MOVEFILE_COPY_ALLOWED           = 0x00000002,
		MOVEFILE_DELAY_UNTIL_REBOOT     = 0x00000004,
		MOVEFILE_WRITE_THROUGH          = 0x00000008,
		MOVEFILE_CREATE_HARDLINK        = 0x00000010,
		MOVEFILE_FAIL_IF_NOT_TRACKABLE  = 0x00000020,
	}
	/***/
	enum MOVEFLAG {
		COPY_ALLOWED          = MOVEFILE_COPY_ALLOWED          , /// 他のボリュームへファイルを移動するよう指示された場合、この関数は、CopyFile 関数と DeleteFile 関数を使って、その動作をシミュレートします。このフラグを、MOVEFILE_DELAY_UNTIL_REBOOT フラグと共に指定することはできません。
		CREATE_HARDLINK       = MOVEFILE_CREATE_HARDLINK       , /// 将来の使用に備えて予約されています。
		DELAY_UNTIL_REBOOT    = MOVEFILE_DELAY_UNTIL_REBOOT    , /// この関数は、オペレーティングシステムを再起動するまでファイルを移動しません。システムは、AUTOCHK（Windows NT/2000 に付属していて、Chkdsk に相当し、起動時に実行できます）を実行した直後、かつページングファイルを作成する前にファイルを移動します。その後、この値が指定されている場合は、この関数は前回のスタートアップに使ったページングファイルを削除できます。このフラグを指定できるのは、Administrators グループに所属するユーザーのコンテキスト、または LocalSystem アカウントのコンテキストだけです。このフラグを、MOVEFILE_COPY_ALLOWED フラグと共に指定することはできません。
		FAIL_IF_NOT_TRACKABLE = MOVEFILE_FAIL_IF_NOT_TRACKABLE , /// Windows 2000：移動元ファイルがリンクソースの場合、このファイルは失敗します。そして、移動後は、そのファイルのリンクトラッキングを行えません。FAT ファイルシステムでフォーマットされたボリュームを移動先として指定すると、このような問題が発生することがあります。
		REPLACE_EXISTING      = MOVEFILE_REPLACE_EXISTING      , /// lpNewFileName パラメータで指定されたファイル名が既に存在していた場合、その内容を lpExistingFileName パラメータで指定されたファイルに置き換えます。lpNewFileName と lpExistingFileName どちらかのパラメータでディレクトリを指定した場合は、この値を利用できません。
		WRITE_THROUGH         = MOVEFILE_WRITE_THROUGH         , /// ファイルがディスク上で実際に移動するまで、制御を返しません。このフラグを指定すると、コピーと削除の操作からなる移動操作をディスクへフラッシュした（実際に書き込んだ）後、この関数から制御が返ることが保証されます。コピー操作が終わった後、フラッシュが実施されます。このフラグは、MOVEFILE_DELAY_UNTIL_REBOOT フラグが指定されていると機能しません。
	}
	/***/
	bool move(in Text SrcPath, in Text NewPath) {
		return cast(bool)MoveFile(SrcPath.ptr, NewPath.ptr);
	}
	/// ditto
	bool move(in Text SrcPath, in Text NewPath, MOVEFLAG MoveFlags) {
		return cast(bool)MoveFileEx(SrcPath.ptr, NewPath.ptr, MoveFlags);
	}
	/***/
	bool copy(in Text SrcPath, in Text NewPath, bool Overwrite) {
		return cast(bool)CopyFile(SrcPath.ptr, NewPath.ptr, !Overwrite);
	}


	private final scope class FileHandle: NeFile {
		this(in Text Path, bool Write=false, string _file=__FILE__, int _line=__LINE__) {
			super(
				Path,
				Write ? ACCESS_MODE.READ | ACCESS_MODE.WRITE: ACCESS_MODE.READ,
				SHARE.ALL,
				null,
				DISPOSITION.OPEN_EXISTING,
				FILE.ATTRIBUTE.NORMAL,
				FLAG.NONE,
				null,
				true
			);
		}
	}

	/***/
	long fileSize(in Text Path) {
		scope Handle=new FileHandle(Path);
		//LARGE_INTEGER LargeInteger;
		LARGE Large;
		enforce(GetFileSizeEx(Handle(), Large.ptr), new NeGuiFileException(ERR.toText));
		return Large.total;
	}
	/***/
	ulong fileSizeCompressed(in Text Path) {
		ERR.initialize;
		DWORD HiSize;
		DWORD LoSize=GetCompressedFileSize(Path.ptr, &HiSize);
		if(LoSize == -1 && ERR.code != ERR.noError) {
			throw new NeGuiFileException(Text("%s, Path -> %s", ERR.toText, Path));
		}
		
		if(HiSize) {
			return cast(ulong)(HiSize * DWORD.max)+1 + LoSize;
		} else {
			return LoSize;
		}
	}

	/***/
	ref TIMESTAMP timeStamp(in Text Path, TIMESTAMP.FLAG Flags=TIMESTAMP.FLAG.ALL) {
		scope Handle=new FileHandle(Path);
		
		FILEDATETIME Create, Access, Update;
		GetFileTime(
			Handle(),
			Flags & TIMESTAMP.FLAG.CREATE ? Create.ptr: null,
			Flags & TIMESTAMP.FLAG.ACCESS ? Access.ptr: null,
			Flags & TIMESTAMP.FLAG.UPDATE ? Update.ptr: null
		);
		auto TimeStamp=new TIMESTAMP;
		with(*TimeStamp) {
			create = Flags & TIMESTAMP.FLAG.CREATE ? new DateTime(DateTime.MODE.STRICT, Create.toLocalFileDateTime): null;
			access = Flags & TIMESTAMP.FLAG.ACCESS ? new DateTime(DateTime.MODE.STRICT, Access.toLocalFileDateTime): null;
			update = Flags & TIMESTAMP.FLAG.UPDATE ? new DateTime(DateTime.MODE.STRICT, Update.toLocalFileDateTime): null;
		}
		TimeStamp.flags = Flags;
		
		return *TimeStamp;
	}
	/***/
	bool timeStamp(in Text Path, const ref TIMESTAMP TimeStamp) {
		scope Handle=new FileHandle(Path, true);
		with(TimeStamp) {
			return cast(bool)SetFileTime(
				Handle(),
				TimeStamp.flags & TIMESTAMP.FLAG.CREATE ? create.toFileDateTime.toSystemFileDateTime.ptr: null,
				TimeStamp.flags & TIMESTAMP.FLAG.ACCESS ? access.toFileDateTime.toSystemFileDateTime.ptr: null,
				TimeStamp.flags & TIMESTAMP.FLAG.UPDATE ? update.toFileDateTime.toSystemFileDateTime.ptr: null
			);
		}
	}

	private Text TempFile(in Text Folder, in Text Prefix, size_t Unique) {
		enforce(isFolder(Folder), new NeGuiFileException(Text("Folder -> %s", Folder)));
		
		auto Buffer=new wchar[PATH.MAX_NAME];
		
		auto Result=GetTempFileName(Folder.ptr, Prefix.ptr, Unique, Buffer.ptr);
		enforce(Result, new NeGuiFileException(ERR.toText));
		
		return Text(Buffer.ptr);
	}
	/***/
	Text makeTempFile(in Text Folder, in Text Prefix=Text.init) {
		return TempFile(Folder, Prefix, 0);
	}
	/***/
	Text getTempFile(in Text Folder, in Text Prefix=Text.init, size_t Unique=1) {
		enforce(Unique, new NeGuiFileException(Text("Unique -> 0")));
		return TempFile(Folder, Prefix, Unique);
	}




	///
	enum BINARY {
		NO    = 0,                /// バイナリじゃない
		BIT32 = SCS_32BIT_BINARY, /// Win32 ベースのアプリケーション
		BIT64 = hide_file.SCS_64BIT_BINARY, /// 
		DOS   = SCS_DOS_BINARY,   /// MS-DOS ベースのアプリケーション
		OS216 = SCS_OS216_BINARY, /// 06 ビット版 OS/2 ベースのアプリケーション
		PIF   = SCS_PIF_BINARY,   /// MS-DOS ベースのアプリケーションを実行する PIF ファイル
		POSIX = SCS_POSIX_BINARY, /// POSIX ベースのアプリケーション
		WOW   = SCS_WOW_BINARY,   /// 16 ビット版 Windows ベースのアプリケーション
	}
	/**
	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	BINARY binaryType(in Text Path) {
		DWORD BinaryTypeSrc;
		enforce(GetBinaryType(Path.ptr, &BinaryTypeSrc), new NeGuiFileException(ERR.toText));
		return cast(BINARY)BinaryTypeSrc;
	}

}

///
struct DISKSPACE {
	/// 
	ULARGE available;
	/// 
	ULARGE total;
	/// 
	ULARGE free;

	/// 
	const ConsText toText() {
		return Text(
			"available -> %s total -> %s free -> %s",
			available, total, free
		);
	}
}

/**

History:
	$(HIST_VER 1.130,
		$(HIST_NEW)
	)
*/
struct VOLUMEINFORMATION {
	///
	Text name;
	///
	DWORD serialNumber;
	///
	size_t maxFileNameLength;
	/***/
	enum FILESYSTEM {
		CASE_IS_PRESERVED       = FS_CASE_IS_PRESERVED,         /// ファイルシステムは、ディスクにファイル名を記録するときに、大文字と小文字を区別して保存します。
		CASE_SENSITIVE          = FS_CASE_SENSITIVE,            /// ファイルシステムは、ファイル名の大文字と小文字の区別をサポートしています。
		UNICODE_STORED_ON_DISK  = FS_UNICODE_STORED_ON_DISK,    /// ファイルシステムは、Unicode のファイル名をサポートしていて、ディスク上でも正しく表示されます。
		PERSISTENT_ACLS         = FS_PERSISTENT_ACLS,           /// ファイルシステムは、ACL（アクセス制御リスト）の保存と適用を行います。たとえば、NTFS は ACL の保存と適用を行いますが、FAT は ACL をサポートしていません。
		FILE_COMPRESSION        = FS_FILE_COMPRESSION,          /// ファイルシステムは、ファイルベースの圧縮をサポートしています。
		VOL_IS_COMPRESSED       = FS_VOL_IS_COMPRESSED,         /// 指定されたボリュームは、圧縮ボリューム（たとえば、DoubleSpace のボリューム）です。
		NAMED_STREAMS           = FILE_NAMED_STREAMS,           /// ファイルシステムは、名前付きストリームをサポートしています。
		SUPPORTS_ENCRYPTION     = FILE_SUPPORTS_ENCRYPTION,     /// ファイルシステムは、暗号化ファイルシステム（EFS）をサポートしています。
		SUPPORTS_OBJECT_IDS     = FILE_SUPPORTS_OBJECT_IDS,     /// ファイルシステムは、オブジェクト識別子をサポートしています。
		SUPPORTS_REPARSE_POINTS = FILE_SUPPORTS_REPARSE_POINTS, /// ファイルシステムは、再解析ポイントをサポートしています。
		SUPPORTS_SPARSE_FILES   = FILE_SUPPORTS_SPARSE_FILES,   /// ファイルシステムは、スパースファイル（疎なファイル、未使用の領域が多い、または同じ値が長く続くファイル）をサポートしています。
		SUPPORTS_VOLUME_QUOTAS  = FILE_VOLUME_QUOTAS,           /// ファイルシステムは、ディスククォータ（ディスク使用量の割り当て機能）をサポートしています。
	}
	///
	FILESYSTEM fileSystemFlags;
	///
	Text fileSystemName;
}




/***/
struct DRIVE {
	static:

	/***/
	ref DISKSPACE diskSpace(in Text Folder) {
		auto DiskSpace=new DISKSPACE;

		enforce(GetDiskFreeSpaceEx(
			Folder.ptr(true),
			DiskSpace.available.ptr,
			DiskSpace.total.ptr,
			DiskSpace.free.ptr
		), new DiskException(ERR.toText));
		
		return *DiskSpace;
	}

	
	/**

	*/
	enum TYPE {
		UNKNOWN     = DRIVE_UNKNOWN,     /// 不明です。 
		NO_ROOT_DIR = DRIVE_NO_ROOT_DIR, /// 指定したルートディレクトリは存在しません。 
		REMOVABLE   = DRIVE_REMOVABLE,   /// ドライブからディスクを抜くことができます。 
		FIXED       = DRIVE_FIXED,       /// ドライブからディスクを抜くことができません。 
		REMOTE      = DRIVE_REMOTE,      /// ネットワーク ドライブです。 
		CDROM       = DRIVE_CDROM,       /// CD-ROM ドライブです。 
		RAMDISK     = DRIVE_RAMDISK,     /// RAM ディスク ドライブです。 
	}
	/***/
	TYPE type(in Text Drive) {
		return cast(TYPE)GetDriveType(Drive.ptr);
	}
	/***/
	BIT!(DWORD) drives() {
		return BIT!(DWORD)(GetLogicalDrives());
	}
	
	/**
	History:
		$(HIST_VER 1.130,
			$(HIST_B 戻り値の内容が不完全。)
			$(HIST_CHG 戻り値の型, Text[], const(Text)[])
		)
	*/
	const(Text)[] names() {
		auto Length=GetLogicalDriveStrings(0, null);
		enforce(Length, new DriveException(ERR.toText));

		auto Buffer=new tchar[Length+1];
		enforce(GetLogicalDriveStrings(Length+1, Buffer.ptr), new DriveException(ERR.toText));

		/+
		auto TextNullList=TEXTNULLLIST(Buffer.ptr);
		return TextNullList.datas;
		+/
		
		auto list=new TextList("\0", "\0\0");
		list.setData(Buffer.ptr);
		return list[];
	}
	version(none)
	debug(file) unittest {
		wl("%(s,) ", names);
	}

	/**
	History:
		$(HIST_VER 1.130,
			$(HIST_NEW)
		)
	*/
	ref VOLUMEINFORMATION volumeInformation(in Text RootPath) {
		auto VolumeNameBuffer=new tchar[MAX_PATH+1];
		DWORD SerialNumber;
		DWORD MaxFileNameLength;
		DWORD FileSystemFlags;
		auto FileSystemName=new tchar[MAX_PATH+1];
		
		enforce(GetVolumeInformation(
			RootPath.ptr,
			VolumeNameBuffer.ptr,
			VolumeNameBuffer.length,
			&SerialNumber,
			&MaxFileNameLength,
			&FileSystemFlags,
			FileSystemName.ptr,
			FileSystemName.length
		),new DriveException(ERR.toText));
		
		auto VolumeInfo=new VOLUMEINFORMATION;
		with(*VolumeInfo) {
			name              = Text(VolumeNameBuffer.ptr);
			serialNumber      = SerialNumber;
			maxFileNameLength = MaxFileNameLength;
			fileSystemFlags   = cast(FILESYSTEM)FileSystemFlags;
			fileSystemName    = Text(FileSystemName.ptr);
		}
		
		return *VolumeInfo;
	}
}


/**
History:
	$(HIST_VER 1.120,
		$(HIST_NEW)
	)
*/
struct FOLDER {
static:
	///
	enum SYSTEM {
		COMPUTER = int.min, /// マイコンピュータ
		CONTROLPANEL,       /// コントロールパネル
		TRASH,              /// ゴミ箱
		NETWORK,            /// マイネットワーク
		NETWORK_CONNECT,    /// ネットワーク接続
		CAMERA_SCANNER,     /// カメラとスキャナ
		PRINTER_FAX,        /// プリンタとFAX
		TASK,               /// タスク
		FONTS,              /// フォント
		ADMINTOOLS,         /// 管理ツール

		_PADDING_VIRTUAL_,

		STARTMENU                = CSIDL_STARTMENU,               /// 「スタートメニュー」（ファイルシステムディレクトリ）
		PROGRAMS                 = CSIDL_PROGRAMS,                /// 「プログラム」（ファイルシステムディレクトリ）
		STARTUP                  = CSIDL_STARTUP,                 /// 「スタートアップ」（ファイルシステムディレクトリ）
		DESKTOPDIRECTORY         = CSIDL_DESKTOPDIRECTORY,        /// 「デスクトップ」上のファイルオブジェクトを格納するフォルダ（ファイルシステムディレクトリ）
		FAVORITES                = CSIDL_FAVORITES,               /// 「お気に入り」（ファイルシステムディレクトリ）
		RECENT                   = CSIDL_RECENT,                  /// 「最近使ったファイル」（ファイルシステムディレクトリ）
		SENDTO                   = CSIDL_SENDTO,                  /// 「SendTo」（ファイルシステムディレクトリ）
		
		PROFILE                  = CSIDL_PROFILE,                 /// Version 5.0 以降： profile フォルダ
		DOCUMENTS                = CSIDL_PERSONAL,                /// 「マイ ドキュメント」（ファイルシステムディレクトリ）
		PICTURES                 = CSIDL_MYPICTURES,              /// Version 5.0 以降： My Pictures フォルダ（ファイルシステムディレクトリ）
		MUSIC                    = hide_file.CSIDL_MYMUSIC,                 /// 「マイ ミュージック」（ファイルシステムディレクトリ）
		VIDEO                    = hide_file.CSIDL_MYVIDEO,                 /// 「マイミュージック」
		
		APPDATA                  = CSIDL_APPDATA,                 /// Version 4.71 以降： 「Application Data」（ファイルシステムディレクトリ）
		APPDATA_LOCAL            = CSIDL_LOCAL_APPDATA,           /// Version 5.0 以降： 「Application Data」（ファイルシステムディレクトリ）
		TEMPLATES                = CSIDL_TEMPLATES,               /// ドキュメントテンプレートが格納されるフォルダ（ファイルシステムディレクトリ）
		
		INTERNET_CACHE           = CSIDL_INTERNET_CACHE,          /// Version 4.72 以降： インターネット一時ファイルを格納するフォルダ（ファイルシステムディレクトリ）
		INTERNET_COOKIES         = CSIDL_COOKIES,                 /// 「Cookies」（ファイルシステムディレクトリ）
		INTERNET_HISTORY         = CSIDL_HISTORY,                 /// 「履歴」（ファイルシステムディレクトリ）
		
		WINDOWS                  = CSIDL_WINDOWS,                 /// Version 5.0 以降： Windows ディレクトリ
		SYSTEM                   = CSIDL_SYSTEM,                  /// Version 5.0 以降： Windows System ディレクトリ
		PROGRAM_FILES            = CSIDL_PROGRAM_FILES,           /// Version 5.0 以降： Program Files フォルダ
		PROGRAM_FILES_COMMON     = CSIDL_PROGRAM_FILES_COMMON,    /// Version 5.0 以降；Windows 2000/XP： Program Files\Common
		
		ALLUSER_STARTMENU        = CSIDL_COMMON_STARTMENU,        /// AllUsers の「スタートメニュー」（ファイルシステムディレクトリ）
		ALLUSER_PROGRAMS         = CSIDL_COMMON_PROGRAMS,         /// AllUsers の「プログラム」（ファイルシステムディレクトリ）
		ALLUSER_STARTUP          = CSIDL_COMMON_STARTUP,          /// AllUsers の「スタートアップ」（ファイルシステムディレクトリ）
		ALLUSER_DESKTOPDIRECTORY = CSIDL_COMMON_DESKTOPDIRECTORY, /// AllUsers の「デスクトップ」（ファイルシステムディレクトリ）
		ALLUSER_FAVORITES        = CSIDL_COMMON_FAVORITES,        /// AllUsers の「お気に入り」（NT系のみ）（ファイルシステムディレクトリ）
		ALLUSER_APPDATA          = CSIDL_COMMON_APPDATA,          /// Version 5.0 以降： AllUsers の「Application Data」（ファイルシステムディレクトリ）
		ALLUSER_TEMPLATES        = CSIDL_COMMON_TEMPLATES,        /// Windows NT/2000/XP： AllUsers のドキュメントテンプレートが格納されるディレクトリ
		ALLUSER_DOCUMENTS        = CSIDL_COMMON_DOCUMENTS,        /// Shfolder.dll： AllUsers のドキュメントテンプレートが格納されるディレクトリ（Windows NT 系および Shfolder.dll がインストールされた Windows 9x）
		ALLUSER_PICTURES         = CSIDL_COMMON_PICTURES,         /// 
		ALLUSER_MUSIC            = CSIDL_COMMON_MUSIC,            /// 
		ALLUSER_VIDEO            = CSIDL_COMMON_VIDEO,            /// 
	}
	///
	private struct VIRTUAL {
		static immutable:
		/// マイコンピュータ
		Text COMPUTER = {data: cast(tchar[])`::{20D04FE0-3AEA-1069-A2D8-08002B30309D}`};
		/// コントロールパネル
		Text CONTROLPANEL = {data: cast(tchar[])`::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\::{21EC2020-3AEA-1069-A2DD-08002B30309D}`};
		/// ゴミ箱
		Text TRASH = {data: cast(tchar[])`::{645FF040-5081-101B-9F08-00AA002F954E}`};
		/// マイネットワーク
		Text NETWORK = {data: cast(tchar[])`::{208D2C60-3AEA-1069-A2D7-08002B30309D}`};
		/// ネットワーク接続
		Text NETWORK_CONNECT = {data: cast(tchar[])`::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\::{21EC2020-3AEA-1069-A2DD-08002B30309D}\::{7007ACC7-3202-11D1-AAD2-00805FC1270E}`};
		/// カメラとスキャナ
		Text CAMERA_SCANNER = {data: cast(tchar[])`::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\::{21EC2020-3AEA-1069-A2DD-08002B30309D}\::{E211B736-43FD-11D1-9EFB-0000F8757FCD}`};
		/// プリンタとFAX
		Text PRINTER_FAX = {data: cast(tchar[])`::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\::{21EC2020-3AEA-1069-A2DD-08002B30309D}\::{2227A280-3AEA-1069-A2DE-08002B30309D}`};
		/// タスク
		Text TASK = {data: cast(tchar[])`::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\::{21EC2020-3AEA-1069-A2DD-08002B30309D}\::{D6277990-4C6A-11CF-8D87-00AA0060F5BF}`};
		/// フォント
		Text FONTS = {data: cast(tchar[])`::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\::{21EC2020-3AEA-1069-A2DD-08002B30309D}\::{D20EA4E1-3957-11d2-A40B-0C5020524152}`};
		/// 管理ツール
		Text ADMINTOOLS = {data: cast(tchar[])`::{20D04FE0-3AEA-1069-A2D8-08002B30309D}\::{21EC2020-3AEA-1069-A2DD-08002B30309D}\::{D20EA4E1-3957-11d2-A40B-0C5020524153}`};
	}
	/***/
	bool isVirtual(const ref Text Path) {
		immutable static Text HEAD = {data: cast(tchar[])`::{`};
		immutable static Text TAIL = {data: cast(tchar[])`}`};
		return Path.startsWith(HEAD) && Path.endsWith(TAIL);
	}
	debug(file) unittest {
		assert(isVirtual(Text("::{}")));
		assert(!isVirtual(Text(":{}")));
		assert(!isVirtual(Text("::{")));
		assert(!isVirtual(Text("}")));
	}
	/***/
	Text virtualPath(SYSTEM Folder)
	in {
		assert(Folder < SYSTEM._PADDING_VIRTUAL_);
	}
	body {
		final switch(cast(int)Folder) {
			case SYSTEM.COMPUTER:
				return VIRTUAL.COMPUTER.dup;
			case SYSTEM.CONTROLPANEL:
				return VIRTUAL.CONTROLPANEL.dup;
			case SYSTEM.TRASH:
				return VIRTUAL.TRASH.dup;
			case SYSTEM.NETWORK:
				return VIRTUAL.NETWORK.dup;
			case SYSTEM.NETWORK_CONNECT:
				return VIRTUAL.NETWORK_CONNECT.dup;
			case SYSTEM.CAMERA_SCANNER:
				return VIRTUAL.CAMERA_SCANNER.dup;
			case SYSTEM.PRINTER_FAX:
				return VIRTUAL.PRINTER_FAX.dup;
			case SYSTEM.TASK:
				return VIRTUAL.TASK.dup;
			case SYSTEM.FONTS:
				return VIRTUAL.FONTS.dup;
			case SYSTEM.ADMINTOOLS:
				return VIRTUAL.ADMINTOOLS.dup;
			
			//default:
			//	assert(false);
		}
	}

	/***/
	Text systemPath(SYSTEM Folder, size_t PathSize=PATH.MAX_PATH)
	in {
		assert(Folder != SYSTEM._PADDING_VIRTUAL_);
		assert(PathSize);
	}
	out(r) {
		if(Folder < SYSTEM._PADDING_VIRTUAL_) {
			assert(isVirtual(r));
		}
	}
	body {
		if(Folder < SYSTEM._PADDING_VIRTUAL_) {
			return virtualPath(Folder);
		} else {
			scope malloc=new Malloc();
			LPITEMIDLIST pidl;
			scope(exit) malloc.free(pidl);
			
			enforce(S_OK == SHGetSpecialFolderLocation(null, Folder, &pidl), new NeGuiFolderException(ERR.toText));
			auto RawPath=new wchar[PathSize];
			enforce(SHGetPathFromIDList(pidl, RawPath.ptr), new NeGuiFolderException(ERR.toText));
			
			return PATH.addFolderSep(Text(RawPath.ptr));
		}
	}
}

	
