﻿/**
RAIIをやりやすく。

気が付けば解放保証人みたいな扱い。
そしてそれすらにもなっていないことに気付く。

リソース管理とデストラクタとGCのメモリ回収がごっちゃになってるんだなぁ。

Bugs:
	最早module名が内容とあっていない
*/
module etc.negui.system.raii;

debug import std.stdio: wl = writefln;
debug(raii) void main() {}

import std.string;
import std.exception;

public import win32.winnt;
public import win32.windows;

public import etc.negui.system.text;
public import etc.negui.system.exception;
public import etc.negui.system.error;

/**

History:
	$(HIST_VER 1.110,
		$(HIST_S 暫定的に継承からtypedefへ。)
	)
*/
/+
class RaiiException: NeGuiException {
	mixin MixInNeGuiException;
}
+/
typedef NeGuiException RaiiException;
/**
RAII基盤

Killメソッドを実装すれば後は自殺するだけさ。

あくまでデストラクタ時にKill()しようとするだけであってGCに一任するものじゃないよ。

History:
	$(HIST_VER 1.00β11,
		$(HIST_MSG 使用していなかったログ機能の撤廃。)
	)
*/
abstract class Raii {
	/**
	デスタラクタ時に自殺を行うフラグ。
	*/
	protected bool Suicide;

	this(bool Suicide) {
		this.Suicide = Suicide;
	}

	/**
	自殺内容。
	
	Suicideの内容は影響しない。

	Throws:
		失敗時に例外を投げる。
	*/
	abstract protected void Kill();

	/**
	手軽に自殺。
	呼び出した後にRaii.Suicideはfalseに設定される。
	Suicideの内容は判定されない。

	Returns:
		自殺と判定された場合はtrue、自殺に見せかけた他殺とか自殺未遂ならfalseを返す。
	*/
	final bool kill() {
		try {
			Kill();
			Suicide = false;
			return true;
		} catch(Throwable e) {
			return false;
		}
	}

	/// さぁ逝こう！
	final ~this() {
		if(Suicide) {
			Kill();
		}
	}

	/**
	身辺整理フラグ。

	Returns:
		準備できているならtrue、まだならfalseを返す。
	*/
	final const bool suicide() {
		return Suicide;
	}
}
/**
Raiiを強制。
*/
interface IRaii {
	/**
	自殺内容。
	Suicideの内容は影響しない。

	Throws:
		失敗時に例外を投げる。
	*/
	protected void Kill();
}


/**
リソースハンドル削除用テンプレート。
限られた用途に限り有効。

Params:
	ResourceName = リソース名＠直接は関係なし。

	ResourceHandle = 解放すべきリソースハンドル。

	KillFunc = ResourceHandleを引数として受け取る処理。
	           成功すると(未実装のKillFuncFailで判定)ResourceHandleはResourceHandle.initが代入される。

	KillFuncFail = KillFuncの失敗値。
*/
mixin template MixInKillResource(alias ResourceName, alias ResourceHandle, alias KillFunc, alias KillFuncFailValue) {
	override void Kill() {
		if(ResourceHandle) {
			if(KillFunc(ResourceHandle) != KillFuncFailValue) {
				ResourceHandle = ResourceHandle.init;
			} else {
				throw new RaiiException(Text("%s.Kill(%s)[%08x]", this.classinfo.name, ResourceName, toHash()));
			}
		}
	}
}
alias MixInKillResource KillResource;

interface IHandle {
	const HANDLE opCall();

	enum DUPOPTION {
		CLOSE_SOURCE = DUPLICATE_CLOSE_SOURCE, /// 複製元ハンドルを閉じます。どのようなエラーステータスが返された場合でもこのハンドルを閉じます。
		SAME_ACCESS  = DUPLICATE_SAME_ACCESS,  /// dwDesiredAccess パラメータを無視します。複製先ハンドルは、複製元ハンドルと同じアクセス権を備えています。
	}
	const protected IHandle Duplicate(HANDLE SourceProcessHandle, HANDLE TargetProcessHandle, DWORD AccessFlags, bool Inheritance, DUPOPTION DupOption);
	/**
	Duplicateの簡易呼び出し。
	*/
	@property const IHandle dup();
}

/**
History:
	$(HIST_VER 1.120,
		$(HIST_NEW)
	)
*/
interface IWait {
	///
	enum WAIT {
		FAILED        = WAIT_FAILED,        /// 
		OBJECT_0      = WAIT_OBJECT_0,      /// 指定したオブジェクトがシグナル状態になったことを意味します。
		ABANDONED     = WAIT_ABANDONED,     /// 指定されたオブジェクトは、放棄されたミューテックスオブジェクトでした（あるスレッドが所有権を持っていましたが、そのスレッドは所有権を解放しないで終了しました）。この関数を呼び出した結果、そのミューテックスオブジェクトの所有権は呼び出し側スレッドに移り、そのミューテックスは非シグナル状態に設定されました。
		IO_COMPLETION = WAIT_IO_COMPLETION, /// 1 つまたは複数の I/O 完了ルーチンが、実行の目的でキューに置かれています。
		TIMEOUT       = WAIT_TIMEOUT,       /// タイムアウト時間が経過し、指定されたオブジェクトが非シグナル状態であったことを意味します。
	}
	/**
		Params:
			TimeOutMs = タイムアウト時間を、ミリ秒（ms）単位で指定します。
			            タイムアウト時間が経過すると、オブジェクトが非シグナル状態であっても、制御を返します。
			            0 を指定すると、この関数は指定されたオブジェクトの状態を調べ、即座に制御を返します。
			            INFINITE を指定すると、オブジェクトがシグナル状態になるまで待機し続けます。 
	*/
	const WAIT wait(size_t TimeOutMs);
}

///
class HandleRaiiException: RaiiException {
	mixin MixInNeGuiException;
}
mixin template MixInHandleRaiiNotDuplicate() {
	override const protected typeof(this) Duplicate(HANDLE SourceProcessHandle, HANDLE TargetProcessHandle, DWORD AccessFlags, bool Inheritance, DUPOPTION DupOption) {
		throw new HandleRaiiException(Text("%s.Duplicate() is not implementation.", this.classinfo.name));
	}
}

mixin template MixInHandleRaiiNotDup() {
	override @property const typeof(this) dup() {
		throw new HandleRaiiException(Text("%s.dup() is not implementation.", this.classinfo.name));
	}
}


/**
ハンドルオブジェクト専用RAII。

Windowsで大活躍のHANDLEをサポートするけど作ったのがnemuxi.drawを作った後だったので足枷。
*/
class HandleRaii: Raii, IHandle {
	/// ハンドル。
	protected HANDLE Handle;

	/**
	コンストラクタ。

	Params:
		Handle = 設定するハンドル。

		Suicide = 死亡フラグ。
	*/
	this(HANDLE Handle, bool Suicide) {
		this.Handle  = Handle;
		super(Suicide);
	}

	protected override void Kill() {
		enforce(CloseHandle(Handle), new HandleRaiiException(ERR.toText));
	}

	
	override const HANDLE opCall() {
		return Handle;
	}
	
	override const protected typeof(this) Duplicate(HANDLE SourceProcessHandle, HANDLE TargetProcessHandle, DWORD AccessFlags, bool Inheritance, DUPOPTION DupOption) {
		HANDLE DupHandle;
		
		auto Result=DuplicateHandle(
			SourceProcessHandle, Handle,
			TargetProcessHandle, &DupHandle,
			AccessFlags, Inheritance, DupOption
		);

		enforce(Result, new HandleRaiiException(ERR.toText));

		return new typeof(this)(DupHandle, Suicide);
	}
	/+
	override @property const HandleRaii dup() {
		throw new HandleRaiiException(Text("%s.dup() is not implementation.", this.classinfo.name));
	}
	+/
	mixin MixInHandleRaiiNotDup;
}

/**
HandleRaiiのコンストラクタ統一。

* ----------------------------------------
* class SuperHandle: HandleRaii, IRaii {
* 	mixin HandleRaiiClass;
* 	
* 	override void Kill() {
* 		del(Handle);
* 	}
* }
* class ExHandle: SuperHandle {
* 	mixin HandleRaiiClass;
* }
*
* auto ex = ExHandle(handle, true);
* ----------------------------------------
*/
mixin template MixInHandleRaiiClass() {
	override this(HANDLE Handle, bool Suicide) {
		super(Handle, Suicide);
	}
}
alias MixInHandleRaiiClass HandleRaiiClass;



/**
初期化が必要。
*/
interface IInitialize {
	/// 初期化の実行。
	void initialize();
}


