﻿/**
History:
	$(HIST_VER 1.130,
		$(HIST_C module, etc.negui.system._hide.text, etc.negui.system.text_hide)
	)
*/
module etc.negui.system.text_hide;

debug import std.stdio: wl = writefln;
debug(text_hide) void main() {}


import std.c.string;

import std.array;
import std.algorithm;
import std.traits;
import std.exception;
import std.uni;
import std.utf;
static import std.string;
import std.windows.charset;

//import hide_exception=etc.negui.system._hide.exception;
import hide_exception=etc.negui.system.exception_hide;
//import etc.negui.system.text;

///
@safe bool _D_ANSI(string s) {
	try {
		validate(s);
		return true;
	} catch(Throwable o) {
		return false;
	}
}
///
string toANSI(string s)
in {
	assert(_D_ANSI(s));
}
body {
	auto p=toMBSz(s);
	auto len=strlen(p);
	return cast(string)p[0 .. len];
}
///
string toD(string s)
out(r) {
	assert(_D_ANSI(r));
}
body {
	return fromMBSz(s.ptr);
}

private immutable {
	auto NUMERAL_ZERO_CHAR=
		"0123456789"
		"ABCDEFGHIJKLM"
		"NOPQRSTUVWXYZ"
	;
	auto NUMERAL_ZERO_NUM = [
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
		11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
		24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36
	];
	static assert(NUMERAL_ZERO_CHAR.length == NUMERAL_ZERO_NUM.length);
}
/**
文字からn進数取得。

Params:
	c = n進数の最大値を表す文字。
	[0-9A-Za-z]。

Returns:
	n進数。

Throws:
	失敗時にNeGuiBaseException。
*/
int numeralChatToNum(char c, string _file=__FILE__, int _line=__LINE__) {
	int n;
	if('0' <= c && c <= '9') {
		n = c - '0';
	} else {
		n = std.string.indexOf(NUMERAL_ZERO_CHAR, c, std.string.CaseSensitive.no);
	}
	enforce(n >= 0, new hide_exception.NeGuiBaseException(std.string.format("c -> %s in [%s]", c, NUMERAL_ZERO_CHAR), null, _file, _line));
	
	return NUMERAL_ZERO_NUM[n];
}





/**
きちんと試験してないから動くか分からん。
*/
private void T_ToCharConv(C, T)(ref C[] s, T fd)
if(isSomeChar!(C) && is(T == delegate) || is(T == return))
in {
	assert(fd);
}
body {
	static if(is(C == char)) {
		C[4] Buffer;
	} else static if(is(C == wchar)) {
		C[2] Buffer;
	}
	
	for(auto i=0; i < s.length; i++) {
		immutable c=fd(s[i]);
		static if(is(C == dchar)) {
			s[i] = c;
		} else {
			immutable Length=encode(Buffer, c);
			//assert(cast(tchar)c == c, "辻褄合わせ未実装");
			if(Length == 1) {
				s[i] = cast(C)c;
			} else {
				s[i .. i+1] = Buffer;
			}
		}
	}
}

void ToCharConv(C)(ref C[] s, dchar function(dchar) fn) {
	T_ToCharConv(s, fn);
}
void ToCharConv(C)(ref C[] s, dchar delegate(dchar) dg) {
	T_ToCharConv(s, dg);
}



/**
大文字小文字を区別する
*/
bool DistinguishMatch(dchar c1, dchar c2) {
	return c1 == c2;
}
/**
大文字小文字を区別しない
*/
bool IgnoreMatch(dchar c1, dchar c2) {
	if(isUniAlpha(c1) && isUniAlpha(c2)) {
		return toUniUpper(c1) == toUniUpper(c2);
	} else {
		return DistinguishMatch(c1, c2);
	}
}

/**
大文字・小文字の区別。
*/
enum _CASE_SENSITIVE {
	YES, /// 区別する。
	NO,  /// 区別しない。
}
/**
開始点。
*/
enum _START {
	HEAD, /// 先頭から。
	TAIL, /// 末尾がら。
}


///
pure nothrow bool function(dchar, dchar) GetCaceSensitiveFunc(_CASE_SENSITIVE CaseSensitive) {
	final switch(CaseSensitive) {
		case _CASE_SENSITIVE.YES: return &DistinguishMatch;
		case _CASE_SENSITIVE.NO:  return &IgnoreMatch;
	}
}


/**
History:
	$(HIST_VER 1.130,
		$(HIST_CHG 戻り値の型, int sizediff_t)
	)
	$(HIST_VER 1.100,
		$(HIST_F 文字列に対応。)
		$(HIST_B std.string.lastIndexOfの挙動と違う。)
		$(HIST_S 引数変更。)
	)
*/
static sizediff_t Find(C)(in C[] SrcWord, in C[] SrcSub, _CASE_SENSITIVE CaseSensitive, _START Start)
if(isSomeChar!(C))
body {
	if(!SrcSub.length || SrcSub.length != 1 && SrcWord.length < SrcSub.length) {
		return -1;
	}

	auto Match=GetCaceSensitiveFunc(CaseSensitive);
	immutable(C)[] Word, Sub;
	final switch(Start) {
		case _START.HEAD:
			Word = SrcWord.idup;
			Sub  = SrcSub.idup;
			break;
		case _START.TAIL:
			Word = SrcWord.idup;
			Sub  = SrcSub.idup;
			Word.reverse;
			Sub.reverse;
			break;
	}

	immutable MaxCount=Sub.length;
	bool Hit=true;
	for(auto i=0, j=0; i < Word.length || !Hit && Word.length < MaxCount; i++) {
		Hit = Match(Word[i], Sub[j]);
		if(Hit) {
			if(++j == MaxCount) {
				if(Start == _START.HEAD) {
					return i - (j - 1);
				} else {
					return Word.length - i - 1;
				}
			}
		} else {
			j = 0;
		}
	}

	return -1;
}






enum FILTERING {
	FILTER,
	EXTRACT,
}
C[] Filtering(C)(in C[] Src, in C[] Characters, FILTERING filtering)
if(isSomeChar!(C))
in {
	assert(Characters.length);
}
body {
	size_t i, j;
	auto Result=new C[Src.length];

	auto Exp = filtering == FILTERING.FILTER
		? (C a, C b) { return a == b;}
		: (C a, C b) { return a != b;}
	;

	while(j < Result.length) {
		immutable Target=Src[j];

		bool Hit=false;
		foreach(c; Characters) {
			if(Exp(c, Target)) {
				Hit = true;
				break;
			}
		}

		if(Hit) {
			j++;
		} else {
			Result[i++] = Src[j++];
		}
	}

	return Result[0 .. i];
}

/**
*/
immutable(C)[] ReplaceToWord(alias HitReplace, C, HEAD)(immutable(C)[] Source, in HEAD Head, C Tail, C Escape)
if(isSomeChar!(C) && (isSomeChar!(HEAD) || is(immutable(C)[] == HEAD)))
{
	if(!Source.length) {
		return null;
	}
	
	immutable(C)[] data;
	auto app=appender(&data);
	size_t i;
	bool Changed=false;
	bool FirstChange=true;

	static if(isSomeChar!(HEAD)) {
		enum HEADLENGTH = 1;
	} else {
		enum HEADLENGTH = Head.length;
	}
	assert(HEADLENGTH);
	void put(T)(in T Data) {
		if(Changed) {
			if(FirstChange) {
				//data = Source[0 .. i];
				app.put(Source[0 .. i]);
				FirstChange = false;
			}
			//data ~= Data;
			app.put(Data);
		}
	}
	bool HeadCheck(immutable(C)[] Target) {
		static if(isSomeChar!(HEAD)) {
			return Target[0] == Head;
		} else {
			return Target.length >= HEADLENGTH && Target[0 .. HEADLENGTH] == Head;
		}
	}
	
	for(; i < Source.length; i++) {
		immutable Next=i + 1 < Source.length;
		
		// エスケープシーケンス？
		if(Next && Source[i] == Escape && HeadCheck(Source[i + 1 .. $])) {
			Changed = true;
			put(Head);
			i+=HEADLENGTH;
			continue;
		}
		if(HeadCheck(Source[i .. $])) {
			enforce(Next, new hide_exception.NeGuiBaseException(std.string.format("last is Head, %s <- %s", Source, Head)));
			// 範囲検索
			auto FindIndex=std.string.indexOf(Source[i + HEADLENGTH .. $], Tail, std.string.CaseSensitive.yes);
			// 置き換え終了文字見つかんない
			enforce(FindIndex >= 0, new hide_exception.NeGuiBaseException(std.string.format("not found Tail, %s -> %s -< %s", Source, Source[i + HEADLENGTH .. $], Tail)));
			
			// それっぽいの見つかった
			bool IsHitReplaceChanged;
			auto ReplaceData=HitReplace(Source[i + HEADLENGTH .. i + HEADLENGTH + FindIndex], IsHitReplaceChanged);
			if(IsHitReplaceChanged) {
				// 元文字列から変換していればそれを設定
				Changed = true;
				put(ReplaceData);
				i += HEADLENGTH + FindIndex;
				continue;
			} else {
				// 元文字列から変換してなければ範囲検索文字まで含めて設定
				// 要はその部分を残す
				put(Source[i .. i + HEADLENGTH + FindIndex]);
				i += FindIndex;
				continue;
			}
		} else {
			put(Source[i]);
		}
	}
	
	if(Changed) {
		return data;
	} else {
		return Source;
	}
}
unittest {
	string rep1(string Word, out bool IsChanged) {
		IsChanged = true;
		return (Word~'@');
	}
	//ReplaceToWord!(rep)("a[bbc".dup, '[', ']', '[');
	assert(ReplaceToWord!(rep1)("a[b]b[[c]", '[', ']', '[') == "ab@b[c]");
	assert(ReplaceToWord!(rep1)("a$b$b$$c", '$', '$', '$') == "ab@b$c");
	assert(ReplaceToWord!(rep1)("[]", '[', ']', '[') == "@");
	assert(ReplaceToWord!(rep1)("[YYYY]-[MM]-[DD]T[hh]:[mm]:[ss].[ff]", '[', ']', '[') == "YYYY@-MM@-DD@Thh@:mm@:ss@.ff@");
	
	string rep2(string Word, out bool IsChanged) {
		IsChanged = false;
		return null;
	}
	assert(ReplaceToWord!(rep2)("a[b]b[[c]", '[', ']', '[') == "a[b]b[c]");
	assert(ReplaceToWord!(rep2)("a[b]b[[c]", '[', ']', '[') == "a[b]b[c]");
	assert(ReplaceToWord!(rep2)("[YYYY]-[MM]-[DD]T[hh]:[mm]:[ss].[ff]", '[', ']', '[') == "[YYYY]-[MM]-[DD]T[hh]:[mm]:[ss].[ff]");
	
	string rep3(string Word, out bool IsChanged) {
		IsChanged = true;
		return "@@>@@";
	}
	assert(ReplaceToWord!(rep3)("a[b]b[[c]", '[', ']', '[') == "a@@>@@b[c]");
	assert(ReplaceToWord!(rep3)("a(((b)b@(((c)", "(((", ')', '@') == "a@@>@@b(((c)");//)))))))))
	
	
	string rep4(string Word, out bool IsChanged) {
		IsChanged = false;
		return null;
	}
	assert(ReplaceToWord!(rep4)("[a][b][c]", '[', ']', '[') == "[a][b][c]");

	string rep5(string Word, out bool IsChanged) {
		IsChanged = Word != "hit";
		return Word;
	}
	assert(ReplaceToWord!(rep5)("[abc]", '[', ']', '[') == "abc");
	assert(ReplaceToWord!(rep5)("[hit]", '[', ']', '[') == "[hit]");
	assert(ReplaceToWord!(rep5)("[hit]hit", '[', ']', '[') == "[hit]hit");
	assert(ReplaceToWord!(rep5)("[][hit]", '[', ']', '[') == "[hit]");
	assert(ReplaceToWord!(rep5)("[unhit][hit]", '[', ']', '[') == "unhit[hit]");
}

/**


ｇｄｇｄになった。
*/
immutable(C)[] ReplaceToEscapeSequence(alias HitReplace, C)(immutable(C)[] Source, C Target)
if(isSomeChar!(C))
{
	if(!Source.length) {
		return null;
	}
	
	immutable(C)[] data;
	auto app=appender(&data);
	bool FirstHit=false;
	size_t i=0;
	size_t LastPos;
	
	for(; i < Source.length; i++) {
		if(Source.length - 1 == i) {
			// すでにラスト
			break;
		}
		
		if(Source[i] == Target) {
			if(!FirstHit) {
				FirstHit = true;
				app.put(Source[0 .. i]);
			} else {
				app.put(Source[LastPos .. i]);
			}

			if(Source[i + 1] == Target) {
				// Escape
				app.put(Target);
				//i+=1;
				//LastPos=i+1;
				LastPos = ++i + 1;
				continue;
			}
			
			size_t ReadLength;
			auto s=HitReplace(Source[i+1 .. $], ReadLength);
			i += ReadLength;
			app.put(s);

			LastPos = i + 1;
		}
	}

	if(FirstHit) {
		if(LastPos) {
			app.put(Source[LastPos .. $]);
		}
		
		return app.data;
	} else {
		return Source;
	}
}
unittest {
	string rep1(string Word, out size_t ReadLength) {
		auto s=Word[0 .. 1];
		ReadLength = s.length;
		return std.string.toupper(s);
	}
	assert(ReplaceToEscapeSequence!(rep1)(":", ':') == ":");
	assert(ReplaceToEscapeSequence!(rep1)("::", ':') == ":");
	assert(ReplaceToEscapeSequence!(rep1)(":::", ':') == "::");
	assert(ReplaceToEscapeSequence!(rep1)("::::", ':') == "::");
	assert(ReplaceToEscapeSequence!(rep1)("abc::a123", ':') == "abc:a123");
	assert(ReplaceToEscapeSequence!(rep1)("abc:a123", ':') == "abcA123");
	assert(ReplaceToEscapeSequence!(rep1)("abc:a123:b", ':') == "abcA123B");

	string a, b;
	a = "abc:d";
	assert(ReplaceToEscapeSequence!(rep1)(a, ':') == "abcD");
	a = "abc:d";
	b = a;
	assert(ReplaceToEscapeSequence!(rep1)(a, '@') is a);
}

enum _REG {
	NONE=0,
	G = 0b_0001,
	I = 0b_0010,
	M = 0b_0100,
}
@safe pure nothrow STRING ToAttributeString(STRING=string)(_REG Attribute)
if(isSomeString!(STRING))
{
	STRING s;
	
	if((Attribute & _REG.G) == _REG.G) {
		s ~= 'g';
	}
	if((Attribute & _REG.I) == _REG.I) {
		s ~= 'i';
	}
	if((Attribute & _REG.M) == _REG.M) {
		s ~= 'm';
	}

	return s;
}
unittest {
	assert(ToAttributeString!(string)(_REG.G) == "g");
	assert(ToAttributeString!(wstring)(_REG.G | _REG.I) == "gi");
	assert(ToAttributeString!(dstring)(_REG.G | _REG.I | _REG.M) == "gim");
	assert(ToAttributeString(_REG.M | _REG.G) == "gm");
}

