﻿/**

*/
module etc.negui.window.menu.ownerdraw;

debug import std.stdio: wl = writefln;
debug(ownerdraw) void main() {}

debug import std.string;

import win32.windows;

import etc.negui.system.base;
import etc.negui.draw.canvas;
import etc.negui.draw.color;
import etc.negui.draw.font;
import etc.negui.draw.brush;
import etc.negui.draw.unit;
import etc.negui.draw.icon;
import etc.negui.draw.unit;
import etc.negui.input.keyboard.keyboard;
import etc.negui.system.type.record;
import etc.negui.negui;
public import etc.negui.window.menu.menu;
public import etc.negui.window.menu.popup;

///
struct DRAWMENUITEM {
	Text text; /// 表示文字列
	Icon icon; /// 表示アイコン
	//wchar key; /// アクセスキー。
	KEY key; /// アクセスキー。

}

/**
オーナードローを少しだけ補助。
描画に必要そうなDRAWMENUITEMをIDと関連付けて保持と共にdwTypeDataにDRAWMENUITEMを設定。

History:
	$(HIST_VER 1.060,
		$(HIST_B たぶんIcon.FIXEDの依存から脱却。)
	)
	$(HIST_VER 1.021,
		$(HIST_P Icon.FIXEDに依存しすぎ。)
	)
*/
class OwnerDrawMenu: PopUp {
	protected {
		/// 対GC。Killの対象注意
		DRAWMENUITEM[COMMAND_ID] DrawItems;

		/// アイコンサイズに対応するピクセルサイズ。
		SIZE IconSize;
	}

	
	/**
	*/
	DRAWMENUITEM[] GetKeyItem(wchar c) {
		DRAWMENUITEM[] items;
		foreach(DrawItem; DrawItems) {
			if(DrawItem.key == c) {
				items ~= DrawItem;
			}
		}
		return items;
	}

	/**
	History:
		$(HIST_VER 1.021,
			$(HIST_NEW)
		)
	*/
	this(const ref SIZE IconSize) {
		this.IconSize = IconSize;
		super();
	}

	/**
	*/
	this(Icon.FIXED IconFixed) {
		this(Icon.fixedToSizeEx(IconFixed));
	}

	/**
	*/
	void iconSize(const ref SIZE IconSize) {
		this.IconSize = IconSize;
	}
	/**
	*/
	const SIZE iconSize() {
		return IconSize;
	}

	/**
	*/
	override void Kill() {
		auto Values=this.DrawItems.values;
		foreach(key; this.DrawItems.keys) {
			this.DrawItems.remove(key);
		}
		foreach(DrawItem; Values) {
			delete DrawItem.icon;
			delete DrawItem.text.data;
		}
		this.DrawItems = null;
			
		super.Kill();
	}

	///
	bool insert(ref const(MENUITEM) MenuItem, DRAWMENUITEM* DrawItem) {
		if(super.insert(MenuItem)) {
			if(DrawItem) {
				DrawItems[MenuItem.id] = *DrawItem;
			}
			
			return true;
		}

		return false;
	}

	/**
	History:
		$(HIST_VER 1.060,
			$(HIST_P 戻り値の型変更。)
		)
	*/
	DRAWMENUITEM* isDrawItem(COMMAND_ID Id) {
		return Id in DrawItems;
	}
	/// 保持しているDRAWMENUITEM参照。
	DRAWMENUITEM* opIndex(COMMAND_ID Id)
	in {
		assert(isDrawItem(Id));
	}
	body {
		return &DrawItems[Id];
	}

	/**
	オーナードロー描画準備。
	*/
	void measureItem(NeGui gui, MEASUREITEM* MeasureItem) {
		assert(this.isDrawItem(MeasureItem.commandID), "設計ミス１");

		immutable IconArea=GetIconControlSize(IconSize);
		
		auto ItemData  =DrawItems[MeasureItem.commandID];

		scope font=new Font(Font.STOCK.MENU);
		scope canvas=gui.canvas();
		canvas.setFont(font);
		SIZE Size=void;
		Size=canvas.getTextArea(ItemData.text);
		MeasureItem.width  = IconArea.cx+ Size.cx;
		MeasureItem.height = GetMenuHeight(IconArea.cy);
	}
	/**
	オーナードロー描画。

	Bugs:
		無効アイテムの描画。
	*/
	void drawItem(DRAWITEM* DrawItem) {
		assert(this.isDrawItem(DrawItem.commandID), "設計ミス２");

		immutable IconArea=GetIconControlSize(IconSize);
		invariant IconWidth =IconArea.cx;
		invariant IconHeight=IconArea.cy;
		
		
		auto ItemData  =DrawItems[DrawItem.commandID];

		SYSTEMCOLOR TextColor;
		SYSTEMCOLOR BackColor;

		if(DrawItem.state & DRAWITEM.STATE.SELECTED) {
			if(DrawItem.state & DRAWITEM.STATE.DISABLED) {
				TextColor = SYSTEMCOLOR.MenuBack();
			} else {
				TextColor = SYSTEMCOLOR.SelectedText();
			}
			BackColor = SYSTEMCOLOR.SelectedBack();
		} else {
			TextColor = SYSTEMCOLOR.MenuText();
			BackColor = SYSTEMCOLOR.MenuBack();
		}
		
		scope back=new Brush(BackColor);
		scope font=new Font(Font.STOCK.MENU);
		scope canvas=DrawItem.canvas;
		canvas.textColor(cast(COLOR)TextColor);
		canvas.backMode(Canvas.BACKMODE.TR);
		canvas.setBrush(back);
		canvas.setFont(font);

		canvas.fillRect(DrawItem.rect);

		auto LineHeight=DrawItem.rect.bottom - DrawItem.rect.top;
		POINT IconPos=void;
		IconPos.x = DrawItem.rect.left + GetSystemMetrics(SM_CXEDGE);
		IconPos.y = DrawItem.rect.top  + LineHeight/2 - IconHeight/2 + GetSystemMetrics(SM_CYEDGE); // あとでかいぜん
		if(ItemData.icon) {
			SIZE Size;
			Size=IconSize;
			canvas.drawIcon(ItemData.icon, IconPos, Size, null);
		}

		IconPos.x += GetIconControlPadding.cx + IconSize.cx;
		IconPos.y =  DrawItem.rect.top  + LineHeight/2 + (PointToPixel(font.height()))/2;
		canvas.drawState(
			ItemData.text,
			IconPos,
			!(DrawItem.state & DRAWITEM.STATE.DISABLED),
			false
		);
	}

	/**
	History:
		$(HIST_VER 1.060,
			$(HIST_S 名前変更。)
		)
	*/
	static enum CLOSE {
		IGNORE  = MNC_IGNORE , /// 押されたキーを破棄し、警告音を鳴らす(デフォルト)
		CLOSE   = MNC_CLOSE  , /// アクティブなメニューをクローズする
		EXECUTE = MNC_EXECUTE, /// 戻り値の下位ワードで指定するメニュー項目を破棄しオーナーウィンドウにWM_COMMANDを送る
		SELECT  = MNC_SELECT , /// 戻り値の下位ワードで指定するメニュー項目を選択する
	}
}


